# web-page: g2.png
import matplotlib.pyplot as plt
import numpy as np
from doc.documentation.directmin import tools_and_data


def read_molecules(filename, molnames):

    with open(filename, 'r') as fd:
        calculated_data_string = fd.read()
        calculated_data = \
            tools_and_data.read_data(calculated_data_string)

    data2return = []
    for _ in molnames:
        data2return.append(_)
        data2return.append(calculated_data[_][0])

    return data2return


f = plt.figure(figsize=(12, 4), dpi=240)
plt.subplot(121)
mollist = \
    ['PH3', 'P2', 'CH3CHO', 'H2COH', 'CS', 'OCHCHO',
     'C3H9C', 'CH3COF', 'CH3CH2OCH3', 'HCOOH']
data = read_molecules('scf-g2-results.txt', mollist)
# scf
x = data[::2]
y = data[1::2]
plt.xticks(range(len(x)), x, rotation=45)
plt.grid(color='k', linestyle=':', linewidth=0.3)
plt.plot(range(len(x)), y, 'b^-', label='SCF', fillstyle='none')

# direct_min
data = read_molecules('dm-g2-results.txt', mollist)
x = data[::2]
# 2 is added to account for the diagonalization
# performed at the beginning and at the end of etdm
y = np.asarray(data[1::2]) + 2

plt.plot(range(len(x)), y, 'ro-', label='ETDM',
         fillstyle='none')
plt.legend()
plt.ylabel('Number of iterations (energy and gradients calls)')

plt.subplot(122)
# direct_min
mollist = \
    ['NO', 'CH', 'OH', 'ClO', 'SH']
data = read_molecules('dm-g2-results.txt', mollist)
x = data[::2]
y = np.asarray(data[1::2]) + 2

plt.xticks(range(len(x)), x, rotation=45)
plt.grid(color='k', linestyle=':', linewidth=0.3)
plt.plot(range(len(x)), y, 'ro-', label='ETDM',
         fillstyle='none')
# scf
data = read_molecules('scf-g2-results.txt', mollist)
x = data[::2]
y = np.asarray(data[1::2])
plt.xticks(range(len(x)), x, rotation=45)
plt.grid(color='k', linestyle=':', linewidth=0.3)
plt.plot(range(len(x)), y, 'bo-', label='SCF',
         fillstyle='none')
plt.legend()
plt.ylabel('Number of iterations (energy and gradients calls)')

f.savefig("g2.png", bbox_inches='tight')



from ase.build import molecule
from ase.io import write
from ase.units import Bohr
from gpaw import GPAW
from gpaw.analyse.hirshfeld import HirshfeldPartitioning

atoms = molecule('H2O')
atoms.center(vacuum=3.5)
atoms.calc = GPAW(h=0.17, txt='h2o.txt')
atoms.get_potential_energy()

# write Hirshfeld charges out
hf = HirshfeldPartitioning(atoms.calc)
for atom, charge in zip(atoms, hf.get_charges()):
    atom.charge = charge
# atoms.write('Hirshfeld.traj') # XXX Trajectory writer needs a fix
atoms.copy().write('Hirshfeld.traj')

# create electron density cube file ready for bader
rho = atoms.calc.get_all_electron_density(gridrefinement=4)
write('density.cube', atoms, data=rho * Bohr**3)



"""Compare Al(fcc) and Al(bcc) at two different plane-wave cutoffs
and two differens k-point densities."""
from ase.build import bulk
from gpaw import GPAW, PW

afcc = 3.985
abcc = 3.190

for kdens in [2.0, 3.0]:
    for ecut in [300, 500]:
        fcc = bulk('Al', 'fcc', a=afcc)
        calc = GPAW(mode=PW(ecut),
                    kpts={'density': kdens},
                    txt=f'bulk-fcc-{ecut:.1f}-{kdens:.1f}.txt')
        fcc.calc = calc
        efcc = fcc.get_potential_energy()

        bcc = bulk('Al', 'bcc', a=abcc)
        calc = GPAW(mode=PW(ecut),
                    kpts={'density': 4.0},
                    txt=f'bulk-bcc-{ecut:.1f}-{kdens:.1f}.txt')
        bcc.calc = calc
        ebcc = bcc.get_potential_energy()

        print(kdens, ecut, efcc, ebcc, efcc - ebcc)



import numpy as np
from ase.build import bulk
from ase.phonons import Phonons
from ase.units import invcm
from gpaw import GPAW
from gpaw.mpi import world

atoms = bulk('C', 'diamond', a=3.567)
calc = GPAW(mode={'name': 'pw', 'ecut': 500},
            kpts=(5, 5, 5), xc='PBE',
            symmetry={'point_group': False},
            convergence={'density': 0.5e-5},
            txt='phonons.txt')

# Phonon calculation
ph = Phonons(atoms, calc, supercell=(1, 1, 1), delta=0.01)
ph.run()

# To display results (optional)
ph.read(method='frederiksen', acoustic=True)
frequencies = ph.band_structure([[0, 0, 0], ])[0]  # Get frequencies at Gamma
if world.rank == 0:
    # save phonon frequencies for later use
    np.save("vib_frequencies.npy", frequencies)
    print('  i    cm^-1')
    print('------------')
    for i, fr in enumerate(frequencies):
        print('{:3d}  {:4.2f}'.format(i, fr / invcm))



# web-page: 2d_eps.png
import matplotlib.pyplot as plt
import numpy as np

plt.figure()

a = np.loadtxt('2d_eps.dat')
plt.plot(a[:, 0], a[:, 1], 'o-', c='b', ms=10, lw=2, label='Bare')
plt.plot(a[:, 0], a[:, 2], 'o-', c='r', ms=8, lw=2, label='Truncated')

plt.xlabel(r'$q$', size=28)
plt.ylabel(r'$\epsilon_{2D}$', size=28)
plt.xticks([0, 10], [r'$\Gamma$', r'$K$'], size=24)
plt.yticks(size=20)
plt.tight_layout()
plt.legend(loc='upper right')
plt.axis([0, a[-1, 0], 0.0, 9.9])

# plt.show()
plt.savefig('2d_eps.png')



import datetime
import sys

import sphinx_rtd_theme
from gpaw import __version__
from gpaw.doctools.aamath import autodoc_process_docstring
try:
    import sphinxcontrib.spelling
except ImportError:
    sphinxcontrib = None

assert sys.version_info >= (3, 6)

sys.path.append('.')

extensions = ['images',
              'ext',
              'sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.extlinks',
              'sphinx.ext.viewcode',
              'sphinx.ext.napoleon',
              'sphinx.ext.mathjax',
              'sphinx.ext.intersphinx']

if sphinxcontrib:
    extensions.append('sphinxcontrib.spelling')
extlinks = {'doi': ('https://doi.org/%s', 'doi: %s'),
            'arxiv': ('https://arxiv.org/abs/%s', 'arXiv: %s'),
            'xkcd': ('https://xkcd.com/%s', 'XKCD: %s')}
spelling_word_list_filename = 'words.txt'
spelling_show_suggestions = True
templates_path = ['templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'GPAW'
copyright = f'{datetime.date.today().year}, GPAW developers'
release = __version__
exclude_patterns = ['build']
default_role = 'math'
pygments_style = 'sphinx'
autoclass_content = 'both'
modindex_common_prefix = ['gpaw.']
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.10', None),
    'ase': ('https://wiki.fysik.dtu.dk/ase', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy', None),
    'pytest': ('https://docs.pytest.org/en/stable', None),
    'mayavi': ('http://docs.enthought.com/mayavi/mayavi', None)}
nitpick_ignore = [('py:class', 'gpaw.calculator.GPAW'),
                  ('py:class', 'gpaw.spinorbit.BZWaveFunctions'),
                  ('py:class', 'GPAW'),
                  ('py:class', 'Atoms'),
                  ('py:class', 'np.ndarray'),
                  ('py:class', 'ase.spectrum.dosdata.GridDOSData'),
                  ('py:class', 'ase.atoms.Atoms'),
                  ('py:class', 'gpaw.point_groups.group.PointGroup'),
                  ('py:class', 'UniformGridFunctions'),
                  ('py:class', 'DomainType'),
                  ('py:class', 'Path'),
                  ('py:class', 'Vector'),
                  ('py:class', 'ArrayLike1D'),
                  ('py:class', 'ArrayLike2D'),
                  ('py:class', 'Array1D'),
                  ('py:class', 'Array2D'),
                  ('py:class', 'Array3D'),
                  ('py:class', 'MPIComm'),
                  ('py:class', 'DomainType'),
                  ('py:class', 'IO')]

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_style = 'gpaw.css'
html_title = 'GPAW'
html_favicon = 'static/gpaw_favicon.ico'
html_static_path = ['static']
html_last_updated_fmt = '%a, %d %b %Y %H:%M:%S'

mathjax3_config = {
    'tex': {
        'macros': {
            'br': '{\\mathbf r}',
            'bk': '{\\mathbf k}',
            'bG': '{\\mathbf G}'}}}

autodoc_typehints = 'description'
autodoc_typehints_description_target = 'documented'


def setup(app):
    app.connect('autodoc-process-docstring',
                lambda app, what, name, obj, options, lines:
                    autodoc_process_docstring(lines))



import pytest
from gpaw.mpi import world
import numpy as np
from ase.build import bulk
from gpaw import GPAW
from gpaw.response.df import DielectricFunction
from gpaw.response.bse import BSE
from gpaw.test import findpeak, equal

pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


@pytest.mark.response
def test_response_bse_aluminum(in_tmp_dir):
    GS = 1
    df = 1
    bse = 1
    check_spectrum = 1

    if GS:
        a = 4.043
        atoms = bulk('Al', 'fcc', a=a)
        calc = GPAW(mode='pw',
                    kpts={'size': (4, 4, 4), 'gamma': True},
                    xc='LDA',
                    nbands=4,
                    convergence={'bands': 'all'})

        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write('Al.gpw', 'all')

    q_c = np.array([0.25, 0.0, 0.0])
    w_w = np.linspace(0, 24, 241)
    eta = 0.2
    ecut = 50
    if bse:
        bse = BSE('Al.gpw',
                  valence_bands=range(4),
                  conduction_bands=range(4),
                  mode='RPA',
                  nbands=4,
                  ecut=ecut,
                  write_h=False,
                  write_v=False,
                  )
        bse_w = bse.get_eels_spectrum(filename=None,
                                      q_c=q_c,
                                      w_w=w_w,
                                      eta=eta)[1]

    if df:
        df = DielectricFunction(calc='Al.gpw',
                                frequencies=w_w,
                                eta=eta,
                                ecut=ecut,
                                hilbert=False)
        df_w = df.get_eels_spectrum(q_c=q_c, filename=None)[1]

    if check_spectrum:
        w_ = 15.1423
        I_ = 25.4359
        wbse, Ibse = findpeak(w_w, bse_w)
        wdf, Idf = findpeak(w_w, df_w)
        equal(wbse, w_, 0.01)
        equal(wdf, w_, 0.01)
        equal(Ibse, I_, 0.1)
        equal(Idf, I_, 0.1)



# web-page: charge-potential.png

import numpy as np
from matplotlib import pyplot
from scipy.stats import linregress

fig, ax = pyplot.subplots()
fig.subplots_adjust(top=0.99, right=0.99)

potentials, charges = [], []
with open('potential.txt', 'r') as f:
    lines = f.read().splitlines()
for line in lines:
    potential, charge = line.split(',')
    potentials += [float(potential)]
    charges += [float(charge)]

line = linregress(charges, potentials)
x_line = [min(charges), max(charges)]
y_line = [line.slope * _ + line.intercept for _ in x_line]

ax.plot(x_line, y_line, '-', color='C1')
ax.plot(charges, potentials, 'o', color='C0')
ax.text(np.mean(x_line) + 0.0 * np.ptp(x_line),
        np.mean(y_line) + 0.1 * np.ptp(y_line),
        '$R^2 = {:.4f}$'.format(line.rvalue**2))

ax.set_xlabel('excess electrons')
ax.set_ylabel('potential, V')
fig.savefig('charge-potential.png')



# creates: g-factors.csv
from math import pi
import ase.units as units
from gpaw.hyperfine import gyromagnetic_ratios

with open('g-factors.csv', 'w') as fd:
    print('Nucleus, g-factor', file=fd)
    for symbol, (n, ratio) in gyromagnetic_ratios.items():
        g = ratio * 1e6 * 4 * pi * units._mp / units._e
        print(f'":math:`^{{{n}}}`\\ {symbol}", {g:.3f}', file=fd)



import numpy as np

from ase.io import write
from gpaw import GPAW
from gpaw.tddft.units import au_to_eV
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix

# Load the objects
calc = GPAW('unocc.gpw', txt=None)
calc.initialize_positions()  # Initialize in order to calculate density
ksd = KohnShamDecomposition(calc, 'ksd.ulm')
dmat = DensityMatrix(calc)
fdm = FrequencyDensityMatrix(calc, dmat, 'fdm.ulm')


def do(w):
    # Select the frequency and the density matrix
    rho_uMM = fdm.FReDrho_wuMM[w]
    freq = fdm.freq_w[w]
    print(f'Frequency: {freq.freq * au_to_eV:.2f} eV')
    print(f'Folding: {freq.folding}')

    # Transform the LCAO density matrix to KS basis
    rho_up = ksd.transform(rho_uMM, broadcast=True)

    # Induced density
    rho_g = ksd.get_density(calc.wfs, [rho_up[0].imag])

    # Save as a cube file
    write(f'ind_{freq.freq * au_to_eV:.2f}.cube', calc.atoms, data=rho_g)

    # Calculate dipole moment for reference
    dm_v = ksd.density.finegd.calculate_dipole_moment(rho_g, center=True)
    absorption = 2 * freq.freq / np.pi * dm_v[0] / au_to_eV * 1e5
    print(f'Total absorption: {absorption:.2f} eV^-1')


do(0)
do(1)



from ase import Atoms
from ase.build import bulk
from gpaw import GPAW
from gpaw import PW
from gpaw.occupations import FermiDirac

cell = bulk('Si', 'fcc', a=5.421).get_cell()
a = Atoms('Si2', cell=cell, pbc=True,
          scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))

calc = GPAW(mode=PW(400),
            xc='PBE',
            occupations=FermiDirac(width=0.001),
            parallel={'domain': 1, 'band': 1},
            kpts={'size': (8, 8, 8), 'gamma': True},
            txt='gs_Si.txt')

a.calc = calc
a.get_potential_energy()

calc.diagonalize_full_hamiltonian(nbands=100)
calc.write('gs_Si.gpw', mode='all')



from ase import Atoms
from gpaw import GPAW
from gpaw.fdtd.poisson_fdtd import FDTDPoissonSolver
from gpaw.fdtd.polarizable_material import (PermittivityPlus,
                                            PolarizableMaterial,
                                            PolarizableSphere)
from gpaw.tddft import TDDFT, DipoleMomentWriter, photoabsorption_spectrum
from gpaw.inducedfield.inducedfield_fdtd import FDTDInducedField
from gpaw.mpi import world
import numpy as np

# Nanosphere radius (Angstroms)
radius = 50.0

# Whole simulation cell (Angstroms)
large_cell = np.array([3 * radius, 3 * radius, 3 * radius])

# Permittivity of Gold
# J. Chem. Phys. 135, 084121 (2011); http://dx.doi.org/10.1063/1.3626549
gold = [[0.2350, 0.1551, 95.62],
        [0.4411, 0.1480, -12.55],
        [0.7603, 1.946, -40.89],
        [1.161, 1.396, 17.22],
        [2.946, 1.183, 15.76],
        [4.161, 1.964, 36.63],
        [5.747, 1.958, 22.55],
        [7.912, 1.361, 81.04]]

# Initialize classical material
classical_material = PolarizableMaterial()

# Classical nanosphere
classical_material.add_component(
    PolarizableSphere(center=0.5 * large_cell,
                      radius=radius,
                      permittivity=PermittivityPlus(data=gold)))

# Poisson solver
poissonsolver = FDTDPoissonSolver(classical_material=classical_material,
                                  cl_spacing=8.0,
                                  qm_spacing=1.0,
                                  cell=large_cell,
                                  communicator=world,
                                  remove_moments=(4, 1))
poissonsolver.set_calculation_mode('iterate')

# Dummy quantum system
atoms = Atoms('H', [0.5 * large_cell], cell=large_cell)
atoms, qm_spacing, gpts = poissonsolver.cut_cell(atoms)
del atoms[:]  # Remove atoms, quantum system is empty

# Initialize GPAW
gs_calc = GPAW(gpts=gpts,
               nbands=-1,
               poissonsolver=poissonsolver)
atoms.calc = gs_calc

# Ground state
energy = atoms.get_potential_energy()

# Save state
gs_calc.write('gs.gpw', 'all')

# Initialize TDDFT and FDTD
kick = [0.001, 0.000, 0.000]
time_step = 10
iterations = 1000

td_calc = TDDFT('gs.gpw')
DipoleMomentWriter(td_calc, 'dm0.dat')

# Attach InducedField to the calculation
frequencies = [2.45]
width = 0.0
ind = FDTDInducedField(paw=td_calc,
                       frequencies=frequencies,
                       width=width)

# Propagate TDDFT and FDTD
td_calc.absorption_kick(kick_strength=kick)
td_calc.propagate(time_step, iterations)

# Save results
td_calc.write('td.gpw', 'all')
ind.write('td.ind')

# Spectrum
photoabsorption_spectrum('dm0.dat', 'spec.dat', width=width)

# Induced field
ind.calculate_induced_field(gridrefinement=2)
ind.write('field.ind', mode='all')



from numpy.polynomial import Polynomial
import ase.units as u
from ase import Atoms
from gpaw import GPAW, PW

A = [3.9, 4.0, 4.1, 4.2]
E = []
for a in A:
    name = f'bulk-fcc-{a:.1f}'
    b = a / 2

    bulk = Atoms('Al',
                 cell=[[0, b, b],
                       [b, 0, b],
                       [b, b, 0]],
                 pbc=True)

    k = 4
    calc = GPAW(mode=PW(300),       # cutoff
                kpts=(k, k, k),     # k-points
                txt=name + '.txt')  # output file

    bulk.calc = calc

    energy = bulk.get_potential_energy()
    calc.write(name + '.gpw')
    E.append(energy)

p = Polynomial.fit(A, E, 3)
a0 = p.deriv(1).roots()[0]
B = p.deriv(2)(a0) * 4 / 9 / a0 / u.J * u.m**3 * 1e-9  # GPa
print((a0, B))

assert abs(a0 - 3.9924) < 0.001
assert abs(B - 87.22) < 0.1



import pytest
import numpy as np
import time

from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.test import findpeak, equal
from gpaw.response.susceptibility import FourComponentSusceptibilityTensor
from gpaw.response.susceptibility import read_component
from gpaw.mpi import size, world


@pytest.mark.kspair
@pytest.mark.response
def test_response_two_aluminum_chi_RPA(in_tmp_dir):
    assert size <= 4**3

    # Ground state calculation

    t1 = time.time()

    a = 4.043
    atoms1 = bulk('Al', 'fcc', a=a)
    atoms2 = atoms1.repeat((2, 1, 1))

    calc1 = GPAW(mode=PW(200),
                 nbands=4,
                 kpts=(8, 8, 8),
                 parallel={'domain': 1},
                 xc='LDA')

    atoms1.calc = calc1
    atoms1.get_potential_energy()

    t2 = time.time()

    calc2 = GPAW(mode=PW(200),
                 nbands=8,
                 kpts=(4, 8, 8),
                 parallel={'domain': 1},
                 xc='LDA')

    atoms2.calc = calc2
    atoms2.get_potential_energy()

    t3 = time.time()

    # Excited state calculation
    q1_qc = [np.array([1 / 8., 0., 0.]), np.array([3 / 8., 0., 0.])]
    q2_qc = [np.array([1 / 4., 0., 0.]), np.array([- 1 / 4., 0., 0.])]
    w = np.linspace(0, 24, 241)

    # Calculate susceptibility using Al
    fcst = FourComponentSusceptibilityTensor(calc1, fxc='RPA',
                                             eta=0.2, ecut=50)
    for q, q_c in enumerate(q1_qc):
        fcst.get_component_array('00', q_c, w, array_ecut=25,
                                 filename='Al1_chiGG_q%d.pckl' % (q + 1))
        world.barrier()

    t4 = time.time()

    # Calculate susceptibility using Al2
    fcst = FourComponentSusceptibilityTensor(calc2, fxc='RPA',
                                             eta=0.2, ecut=50)
    for q, q_c in enumerate(q2_qc):
        fcst.get_component_array('00', q_c, w, array_ecut=25,
                                 filename='Al2_chiGG_q%d.pckl' % (q + 1))
        world.barrier()

    t5 = time.time()

    parprint('')
    parprint('Ground state calc 1 took', (t2 - t1), 'seconds')
    parprint('Ground state calc 2 took', (t3 - t2), 'seconds')
    parprint('Susceptibility calc 1 took', (t4 - t3), 'seconds')
    parprint('Susceptibility calc 2 took', (t5 - t4), 'seconds')

    # Check that results are consistent, when structure is simply repeated

    # Read results
    w11_w, G11_Gc, chiks11_wGG, chi11_wGG = read_component('Al1_chiGG_q1.pckl')
    w21_w, G21_Gc, chiks21_wGG, chi21_wGG = read_component('Al2_chiGG_q1.pckl')
    w12_w, G12_Gc, chiks12_wGG, chi12_wGG = read_component('Al1_chiGG_q2.pckl')
    w22_w, G22_Gc, chiks22_wGG, chi22_wGG = read_component('Al2_chiGG_q2.pckl')

    # Check that reciprocal lattice vectors remain as assumed in check below
    equal(np.linalg.norm(G11_Gc[0]), 0., 1e-6)
    equal(np.linalg.norm(G21_Gc[0]), 0., 1e-6)
    equal(np.linalg.norm(G12_Gc[0]), 0., 1e-6)
    equal(np.linalg.norm(G22_Gc[1] - np.array([1, 0, 0])), 0., 1e-6)

    # Check plasmon peaks remain the same
    wpeak11, Ipeak11 = findpeak(w11_w, -chi11_wGG[:, 0, 0].imag)
    wpeak21, Ipeak21 = findpeak(w21_w, -chi21_wGG[:, 0, 0].imag)
    equal(wpeak11, wpeak21, 0.02)
    equal(Ipeak11, Ipeak21, 1.0)
    wpeak12, Ipeak12 = findpeak(w12_w, -chi12_wGG[:, 0, 0].imag)
    wpeak22, Ipeak22 = findpeak(w22_w, -chi22_wGG[:, 1, 1].imag)
    equal(wpeak12, wpeak22, 0.05)
    equal(Ipeak12, Ipeak22, 1.0)



from ase.io import write
from gpaw import GPAW, Mixer
from ase.build import molecule

CO = molecule('CO')
CO.set_cell((6., 6., 6.))
CO.center()

calc = GPAW(h=0.2,
            nbands=8,
            mixer=Mixer(beta=0.1, nmaxold=5, weight=50.0),
            txt='CO.txt')

CO.calc = calc
CO.get_potential_energy()

# Write wave functions to gpw file
calc.write('CO.gpw', mode='all')

# Generate cube-files of the orbitals.
for n in range(calc.get_number_of_bands()):
    wf = calc.get_pseudo_wave_function(band=n)
    write('CO%d.cube' % n, CO, data=wf)



import numpy as np
from gpaw.directmin.etdm import random_a


class Derivatives:

    def __init__(self, etdm, wfs, c_ref=None, a_vec_u=None,
                 update_c_ref=False, eps=1.0e-7, random_amat=False):
        """
        :param etdm:
        :param wfs:
        :param c_ref: reference orbitals C_ref
        :param a_vec_u: skew-Hermitian matrix A
        :param update_c_ref: if True update reference orbitals
        :param eps: finite difference displacement
        :param random_amat: if True, use random matrix A
        """

        self.eps = eps

        # initialize vectors of elements matrix A
        if a_vec_u is None:
            self.a_vec_u = {u: np.zeros_like(v) for u,
                                                    v in etdm.a_vec_u.items()}

        if random_amat:
            for kpt in wfs.kpt_u:
                u = etdm.kpointval(kpt)
                a = random_a(etdm.a_vec_u[u].shape, wfs.dtype)
                wfs.gd.comm.broadcast(a, 0)
                self.a_vec_u[u] = a

        # initialize orbitals:
        if c_ref is None:
            self.c_ref = etdm.dm_helper.reference_orbitals
        else:
            self.c_ref = c_ref

        # update ref orbitals if needed
        if update_c_ref:
            etdm.rotate_wavefunctions(wfs, self.a_vec_u, etdm.n_dim,
                                      self.c_ref)
            etdm.dm_helper.set_reference_orbitals(wfs, etdm.n_dim)
            self.c_ref = etdm.dm_helper.reference_orbitals
            self.a_vec_u = {u: np.zeros_like(v) for u,
                                                    v in etdm.a_vec_u.items()}

    def get_analytical_derivatives(self, etdm, ham, wfs, dens,
                                   what2calc='gradient'):
        """
           Calculate analytical gradient or approximation to the Hessian
           with respect to the elements of a skew-Hermitian matrix

        :param etdm:
        :param ham:
        :param wfs:
        :param dens:
        :param what2calc: calculate gradient or Hessian
        :return: analytical gradient or Hessian
        """

        assert what2calc in ['gradient', 'hessian']

        if what2calc == 'gradient':
            # calculate analytical gradient
            analytical_der = etdm.get_energy_and_gradients(self.a_vec_u,
                                                           etdm.n_dim,
                                                           ham, wfs, dens,
                                                           self.c_ref)[1]
        else:
            # Calculate analytical approximation to hessian
            analytical_der = np.hstack([etdm.get_hessian(kpt).copy()
                                        for kpt in wfs.kpt_u])
            analytical_der = construct_real_hessian(analytical_der)
            analytical_der = np.diag(analytical_der)

        return analytical_der

    def get_numerical_derivatives(self, etdm, ham, wfs, dens,
                                  what2calc='gradient'):
        """
           Calculate numerical gradient or Hessian with respect to
           the elements of a skew-Hermitian matrix using central finite
           differences

        :param etdm:
        :param ham:
        :param wfs:
        :param dens:
        :param what2calc: calculate gradient or Hessian
        :return: numerical gradient or Hessian
        """

        assert what2calc in ['gradient', 'hessian']

        # total dimensionality if matrices are real
        dim = sum([len(a) for a in self.a_vec_u.values()])
        steps = [1.0, 1.0j] if etdm.dtype == complex else [1.0]
        use_energy_or_gradient = {'gradient': 0, 'hessian': 1}

        matrix_exp = etdm.matrix_exp
        if what2calc == 'gradient':
            numerical_der = {u: np.zeros_like(v) for u,
                                                     v in self.a_vec_u.items()}
        else:
            numerical_der = np.zeros(shape=(len(steps) * dim,
                                            len(steps) * dim))
            # have to use exact gradient when Hessian is calculated
            etdm.matrix_exp = 'egdecomp'

        row = 0
        f = use_energy_or_gradient[what2calc]
        for step in steps:
            for kpt in wfs.kpt_u:
                u = etdm.kpointval(kpt)
                for i in range(len(self.a_vec_u[u])):
                    a = self.a_vec_u[u][i]

                    self.a_vec_u[u][i] = a + step * self.eps
                    fplus = etdm.get_energy_and_gradients(
                        self.a_vec_u, etdm.n_dim, ham, wfs, dens,
                        self.c_ref)[f]

                    self.a_vec_u[u][i] = a - step * self.eps
                    fminus = etdm.get_energy_and_gradients(
                        self.a_vec_u, etdm.n_dim, ham, wfs, dens,
                        self.c_ref)[f]

                    derf = apply_central_finite_difference_approx(
                        fplus, fminus, self.eps)

                    if what2calc == 'gradient':
                        numerical_der[u][i] += step * derf
                    else:
                        numerical_der[row] = construct_real_hessian(derf)

                    row += 1
                    self.a_vec_u[u][i] = a

        if what2calc == 'hessian':
            etdm.matrix_exp = matrix_exp

        return numerical_der


def construct_real_hessian(hess):

    if hess.dtype == complex:
        hess_real = np.hstack((np.real(hess), np.imag(hess)))
    else:
        hess_real = hess

    return hess_real


def apply_central_finite_difference_approx(fplus, fminus, eps):

    if isinstance(fplus, dict) and isinstance(fminus, dict):
        assert (len(fplus) == len(fminus))
        derf = np.hstack([(fplus[k] - fminus[k]) * 0.5 / eps
                          for k in fplus.keys()])
    elif isinstance(fplus, float) and isinstance(fminus, float):
        derf = (fplus - fminus) * 0.5 / eps
    else:
        raise ValueError()

    return derf



import numpy as np
import pytest
from gpaw.core.matrix import Matrix
from gpaw.mpi import broadcast_exception, world


@pytest.mark.parametrize('dtype', [float, complex])
@pytest.mark.skipif(world.size > 2, reason='size>2')
def test_invcholesky(dtype):
    S0 = Matrix(2, 2, dist=(world, 1, 1), dtype=dtype)
    S = S0.new(dist=(world, world.size, 1))
    if world.rank == 0:
        if dtype == float:
            S0.data[:] = np.array([[1.0, 117], [0.1, 2.0]])
        else:
            S0.data[:] = np.array([[1.0, 117], [0.1j, 2.0]])
    S0.redist(S)
    L = S.copy()
    L.invcholesky()
    S.tril2full()
    if world.size == 1:
        print(L.data @ S.data @ L.data.T.conj())
    else:
        print(world.rank, L.data)
    A = L.multiply(S, opa='N').multiply(L, opb='C')
    print(A.data)
    A.redist(S0)
    with broadcast_exception(world):
        if world.rank == 0:
            assert abs(S0.data - np.eye(2)).max() < 1e-14



def check():
    import numpy as np
    from ase.io import read

    configs = read('co.traj@:')
    e = [config.get_potential_energy() for config in configs]
    a = np.array([config.cell[0, 0] for config in configs])
    c = np.array([config.cell[2, 2] for config in configs])
    f = np.array([a**0, a, c, 0.5 * a**2, a * c, 0.5 * c**2])
    p = np.linalg.lstsq(f.T, e, rcond=None)[0]
    p1 = p[1:3]
    p2 = np.array([(p[3], p[4]),
                   (p[4], p[5])])
    a0, c0 = np.linalg.solve(p2.T, -p1)
    assert abs(a0 - 2.4949) < 0.001
    assert abs(c0 - 4.0419) < 0.001
    assert abs(a[4] - a0) < 0.001
    assert abs(c[4] - c0) < 0.005


def workflow():
    from myqueue.workflow import run
    with run(script='co.py', cores=4, tmax='5h'):
        run(function=check)



import warnings

import numpy as np
from gpaw.mpi import MPIComm
from gpaw.new.brillouin import IBZ, BZPoints
from gpaw.rotation import rotation
from gpaw.symmetry import Symmetry as OldSymmetry


def create_symmetries_object(atoms, ids=None, magmoms=None, parameters=None):
    ids = ids or [()] * len(atoms)
    if magmoms is None:
        pass
    elif magmoms.ndim == 1:
        ids = [id + (m,) for id, m in zip(ids, magmoms)]
    else:
        ids = [id + tuple(m) for id, m in zip(ids, magmoms)]
    symmetry = OldSymmetry(ids,
                           atoms.cell.complete(),
                           atoms.pbc,
                           **(parameters or {}))
    symmetry.analyze(atoms.get_scaled_positions())
    return Symmetries(symmetry)


def mat(rot_cc) -> str:
    """Convert 3x3 matrix to str.

    >>> mat([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])
    '[[-1,  0,  0], [ 0,  1,  0], [ 0,  0,  1]]'

    """
    return '[[' + '], ['.join(', '.join(f'{r:2}'
                                        for r in rot_c)
                              for rot_c in rot_cc) + ']]'


class Symmetries:
    """Wrapper for old symmetry object.  Exposes what we need now."""
    def __init__(self, symmetry):
        self.symmetry = symmetry
        self.rotation_scc = symmetry.op_scc
        self.translation_sc = symmetry.ft_sc
        self.a_sa = symmetry.a_sa
        cell_cv = symmetry.cell_cv
        self.rotation_svv = np.einsum('vc, scd, dw -> svw',
                                      np.linalg.inv(cell_cv),
                                      self.rotation_scc,
                                      cell_cv)
        self.rotation_lsmm = [
            np.array([rotation(l, r_vv) for r_vv in self.rotation_svv])
            for l in range(4)]
        self._rotations = {}

    def __len__(self):
        return len(self.rotation_scc)

    def __str__(self):
        lines = ['symmetry:',
                 f'  number of symmetries: {len(self)}']
        if self.symmetry.symmorphic:
            lines.append('  rotations: [')
            for rot_cc in self.rotation_scc:
                lines.append(f'    {mat(rot_cc)},')
        else:
            nt = self.translation_sc.any(1).sum()
            lines.append(f'  number of symmetries with translation: {nt}')
            lines.append('  rotations and translations: [')
            for rot_cc, t_c in zip(self.rotation_scc, self.translation_sc):
                a, b, c = t_c
                lines.append(f'    [{mat(rot_cc)}, '
                             f'[{a:6.3f}, {b:6.3f}, {c:6.3f}]],')
        lines[-1] = lines[-1][:-1] + ']\n'
        return '\n'.join(lines)

    def reduce(self,
               bz: BZPoints,
               comm: MPIComm = None,
               strict: bool = True) -> IBZ:
        (_, weight_k, sym_k, time_reversal_k, bz2ibz_K, ibz2bz_k,
         bz2bz_Ks) = self.symmetry.reduce(bz.kpt_Kc, comm)

        if -1 in bz2bz_Ks:
            msg = 'Note: your k-points are not as symmetric as your crystal!'
            if strict:
                raise ValueError(msg)
            warnings.warn(msg)

        return IBZ(self, bz, ibz2bz_k, bz2ibz_K, weight_k)

    def check_positions(self, fracpos_ac):
        self.symmetry.check(fracpos_ac)

    def symmetrize_forces(self, F_av):
        return self.symmetry.symmetrize_forces(F_av)

    def rotations(self, l_j):
        ells = tuple(l_j)
        rotation_sii = self._rotations.get(ells)
        if rotation_sii is None:
            ni = sum(2 * l + 1 for l in l_j)
            rotation_sii = np.zeros((len(self), ni, ni))
            i1 = 0
            for l in l_j:
                i2 = i1 + 2 * l + 1
                rotation_sii[:, i1:i2, i1:i2] = self.rotation_lsmm[l]
                i1 = i2
            self._rotations[ells] = rotation_sii
        return rotation_sii



from ase.build import fcc100, add_adsorbate
from gpaw import GPAW, PW

slab = fcc100('Al', (2, 2, 2), a=4.05, vacuum=7.5)
add_adsorbate(slab, 'Na', 4.0)
slab.center(axis=2)

slab.calc = GPAW(mode=PW(300),
                 poissonsolver={'dipolelayer': 'xy'},
                 txt='pwcorrected.txt',
                 xc='PBE',
                 setups={'Na': '1'},
                 kpts=(4, 4, 1))
e3 = slab.get_potential_energy()
slab.calc.write('pwcorrected.gpw')



# web-page: bse_Si.png
import matplotlib.pyplot as plt
import numpy as np

plt.figure()

a = np.loadtxt('eps_rpa_Si.csv', delimiter=',')
plt.plot(a[:, 0], a[:, 4], label='RPA', lw=2)

a = np.loadtxt('eps_bse_Si.csv', delimiter=',')
plt.plot(a[:, 0], a[:, 2], label='BSE', lw=2)

plt.xlabel(r'$\hbar\omega\;[eV]$', size=24)
plt.ylabel(r'$\epsilon_2$', size=24)
plt.xticks(size=20)
plt.yticks(size=20)
plt.tight_layout()
plt.axis([2.0, 6.0, None, None])
plt.legend()

# plt.show()
plt.savefig('bse_Si.png')



from myqueue.workflow import run


def workflow():
    runs = [run(script='gaas.py', args=[1], tmax='1h'),
            run(script='gaas.py', args=[2], cores=8, tmax='1h'),
            run(script='gaas.py', args=[3], cores=24, tmax='2h'),
            run(script='gaas.py', args=[4], cores=48, tmax='24h')]
    with run(script='electrostatics.py', cores=8, processes=1, tmax='15m',
             deps=runs):
        run(script='plot_energies.py')
        run(script='plot_potentials.py')

    runs = [run(script='BN.py', args=[1], cores=8, tmax='1h'),
            run(script='BN.py', args=[2], cores=8, tmax='1h'),
            run(script='BN.py', args=[3], cores=8, tmax='2h'),
            run(script='BN.py', args=[4], cores=24, tmax='2h'),
            run(script='BN.py', args=[5], cores=48, tmax='2h')]
    with run(script='electrostatics_BN.py', tmax='15m', deps=runs):
        run(script='plot_energies_BN.py')
        run(script='plot_potentials_BN.py')
        run(script='plot_epsilon.py')



# web-page: traces.png
import pickle
import numpy as np
import ase.io
from ase.data import covalent_radii as radii
from ase.data.colors import jmol_colors
from matplotlib import pyplot
from matplotlib.patches import Circle
from matplotlib.ticker import (MultipleLocator, AutoMinorLocator)


def get_electron_trace(filename):
    with open(filename, 'rb') as f:
        n = pickle.load(f)
    # Multiply each point by its local cell volume to convert from
    # electron density to electrons in each cube.
    vol_adjust = atoms.cell.volume / np.prod(n.shape)
    n *= vol_adjust
    z_trace = np.sum(n, axis=(0, 1))
    z_trace = np.cumsum(z_trace)
    return z_trace


def get_potential_trace(filename):
    with open(filename, 'rb') as f:
        phi = pickle.load(f)
    phi = phi.mean(axis=(0, 1))
    return phi


fig, ax_potential = pyplot.subplots(figsize=(6.4, 2.2))
fig.subplots_adjust(bottom=0.2, top=0.99, left=0.11, right=0.90)
atoms = ase.io.read('atoms.traj')

###########################
# Potential axis.
###########################

z_trace44 = get_potential_trace('esp4.4V.pckl')
z_trace43 = get_potential_trace('esp4.3V.pckl')
z_trace44 -= z_trace44[0]
z_trace43 -= z_trace43[0]
diff = z_trace43 - z_trace44
zs = np.linspace(0., atoms.cell[2][2], num=len(diff))
ax_potential.plot(zs, diff, color='C1', linewidth=2)
label_at = int(len(zs) * 0.80)
ax_potential.text(zs[label_at], diff[label_at] + 0.01, 'potential',
                  ha='center', va='center', color='C1')

###########################
# Electrons axis.
###########################

ax_electrons = ax_potential.twinx()
# ax1.set_zorder(ax2.get_zorder()+1)

z_trace44 = get_electron_trace('all4.4V.pckl')
z_trace43 = get_electron_trace('all4.3V.pckl')
diff = z_trace43 - z_trace44
ax_electrons.plot(zs, diff)
ax_electrons.text(zs[label_at], diff[label_at] - 0.003, 'electrons',
                  ha='center', va='center', color='C0')

###########################
# Atoms axis.
###########################

ax_atoms = ax_potential.twinx()

# Sort atoms by x value.
cell = atoms.cell
atomslist = [atom for atom in atoms]
atomslist = sorted(atomslist, key=lambda atom: atom.x)
atomslist.reverse()

# Add the atoms to the plot as circles.
for atom in atomslist:
    color = jmol_colors[atom.number]
    radius = radii[atom.number]
    circle = Circle((atom.z, atom.y), radius, facecolor='none',
                    edgecolor='0.5', linewidth=0.5)
    ax_atoms.add_patch(circle)

###########################
# Touch up.
###########################

ax_atoms.axis('equal')
ax_potential.set_xlabel(r'$z$ coordinate, $\mathrm{\AA}$')
ax_atoms.set_yticks([])

ax_potential.set_zorder(2)
ax_electrons.set_zorder(1)
ax_potential.patch.set_visible(False)

ax_potential.set_ylabel(r'$\phi_{xy}$(4.4 V) - $\phi_{xy}$(4.3 V), V')
ax_potential.yaxis.set_major_locator(MultipleLocator(0.1))
ax_potential.yaxis.set_minor_locator(AutoMinorLocator(5))
ax_electrons.set_ylabel(r'$n_{xy}$(4.4 V) - $n_{xy}$(4.3 V)')
ax_electrons.yaxis.set_major_locator(MultipleLocator(0.01))
ax_electrons.yaxis.set_minor_locator(AutoMinorLocator(2))

fig.savefig('traces.png')



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Introduction to Python and ASE and some useful libraries

Python is a programming language that has become a de-facto standard within scientific programming.  It is also very popular for teaching programming and computer science. One of the advantages of the language is that it is relatively easy to read and understand pre-existing code.

### Python Notebooks

We are working with Python notebooks.  A notebook is split into cells (this is a text cell, most cells contain a block of Python code).  You can execute a cell by pressing Shift+Enter.  You can edit a text cell by doubleclicking it.  You can change the type of a cell on the menu bar above, use "Markup" for text and "Code" for Python code.  There are also buttons for adding and removing cells.  Cells can be split into two (or merged) on the Edit menu.
"""

# %%
"""
## Importing the necesary modules

A lot of helper packages for e.g. numeric calculations, plotting, atomic-scale simulations etc are available in Python, but you need to "import" the module to get access to it (you also often need to install it, but that has been taken care of here).

A few packages support special integration with Notebooks, this includes the plotting package ``matplotlib``, where the integration is enabled with a notebook-specific command starting with the ``%`` character.

You can add comments to Python code by using the ``#`` character.

Some links for reference:
* [Python](https://www.python.org/)
* Numerical package [NumPy](http://www.numpy.org/)
* Scientific package [Scipy](https://www.scipy.org/)
* Plotting package [Matplotlib](https://matplotlib.org/)
"""

# %%
# magic: %matplotlib notebook
import matplotlib.pyplot as plt    # For nice plotting
import numpy as np                 # Mathematical operations

# %%
"""
Let's try some simple stuff.
"""

# %%
2 + 3

# %%
print('Hello')

# %%
print('Hello ' * 5)

# %%
# help is useful for getting help of af method
help(print)

# you can also use a question mark to get help in a Jupyter notebook,
# this opens up a new window (close it if you like)
# magic: ?print

# %%
"""
## Variables and data types

In python, variables contain data of different types, such as numbers or text strings.  You can print a variable with the ``print()`` function:
"""

# %%
a = 42
mypi = 3.14
b = "some text"
print(a)
print("Pi is approximately", mypi)
print(b)

# %%
"""
The usual mathematical operations are possible.  The operator for exponentiation is the double star.
"""

# %%
print(2 * a)
print(16**2)
print(a + 2 * mypi)

# %%
"""
### A `list` is an ordered collection of arbitrary objects

You will need lists of data. A list of data can be specified at once, or built gradually. The latter is demonstated later.
"""

# %%
primes = [2, 3, 5, 7, 11, 13, 17, 19, 23]
print(primes)

# %%
"""
Lists are indexed starting with 0, so primes[1] is the *second* prime.  You can also access a list from the end, using negative numbers.
"""

# %%
print(primes[1])
print(primes[-1])

# %%
"""
A `list` can contain arbitrary objects
"""

# %%
# a list
l = [1, ('gg', 7), 'hmm', 1.2]
print(l)
print(l[1])   # Python counts from zero, so this is the second element
print(l[-2])  # indexing with negative numbers counts from the end

# %%
"""
### A `dict`  is a mapping from keys to values
"""

# %%
d = {'s': 0, 'p': 1}
print(d)
print(d['p'])
print('Removing an element from the dictionary')
del d['s']
print(d)

# %%
"""
### A `tuple`  is an ordered collection like a list but is *immutable*
useful for keywords in `dict`
"""

# %%
# with a list we can reassign values
x = [2, 3]
x[0] = 100
print(x)
# this it not possible with a tuple
y = (2, 3)
print('y = ', y)
try:
    y[0] = 100
except Exception as x:
    print(x)
print('y = ', y)

# %%
"""
### A ``namedtuple`` is a tuple where the elements also have names.

They can be accessed by index or by name.
"""

# %%
from collections import namedtuple
SKN = namedtuple('IndexSKN', ('spin', 'kpt', 'band'))
a = SKN(0, 3, 4)
print(a)
# The elements may be accessed by index (like a normal tuple) or by name:
print(a[1])
print(a.kpt)

# %%
# lets try and use a namedtuple as keys for a dict
d = {}
d[SKN(0, 10, 5)] = 3.14
d[SKN(0, 1, 3)] = 2.72
print(d)

# %%
key = SKN(spin=0, kpt=1, band=3)
print(d[key])
print(d[(0, 1, 3)])  # one can also use a normal tuple as key

# %%
"""
# NumPy
#### NumPy arrays are heavely used in [ASE](https://wiki.fysik.dtu.dk/ase/)
ASE makes heavy use of an extension to Python called NumPy. The NumPy module
defines an `ndarray` type that can hold large arrays of uniform
multidimensional numeric data. An array is similar to a `list` or a `tuple`,
but it is a lot more powerful and efficient.
"""

# %%
import numpy as np
x = np.array([1, 2, 3])
print(x)
print(x.mean())

# %%
# Multidimensional array
a = np.zeros((3, 2))
a[:, 1] = 1.0
a[1, :] = 2.0
print("The shape of the array:", a.shape)
print("Number of dimensions:", a.ndim)
print("Data type:", a.dtype)
print("And one can print the entire array (if it is not too big):")
print(a)

# %%
# Matrix muliplication
print('shape of a', a.shape)
print('shape of a.T', a.T.shape)  # .T transpose a matrix
b = np.dot(a, a.T)
print(b)
# in a more READABLE way one can use @ to dot matrices together
c = a @ a.T
print('is c equal to b:', np.alltrue(c == b))

# %%
# Elementwise multiplication
d = a * a
print(d)

# %%
# Random Hermitian matrix
rand = np.random.rand
H = rand(6, 6) + 1j * rand(6, 6)  # 1j = sqrt(-1)
H = H + H.T.conj()

# Eigenvalues and eigenvectors
eps, U = np.linalg.eig(H)

#  Make print of numpy arrays less messy:
np.set_printoptions(precision=3, suppress=True)
print('The eigenvalues are:', eps.real)

# lets try and sort them
sorted_indices = eps.real.argsort()
eps = eps[sorted_indices]
U = U[:, sorted_indices]
print('after sorting: ', eps.real)

# Check that U diagonalizes H
D1 = np.diag(eps)  # Diagonal matrix
D2 = U.T.conj() @ H @ U  # Diagonalized H matrix
print('Diagonal matrix (from eigenvalues):')
print(D1)
print('Diagonal matrix (tranforming H):')
print(D2)
print('Are the numbers in the two matrices close to each other?')
print(np.allclose(D2, D1))

# %%
"""

# Plotting with matplotlib
 see here for more details [Matplotlib](https://matplotlib.org/)
"""

# %%
# In a Jupyter Notebook, this magic line gives nice inline figures,
# with interactive possibilities.
# This line MUST appear before you import matplotlib or a package
# using matplotlib (e.g. ase)
# magic: %matplotlib notebook
import matplotlib.pyplot as plt
import numpy as np

# %%
# Some data
x = [1.4, 2.0, 3.0, 3.3]
y = [10, -1.1, 2.2, 5.0]
y2 = [1, 2, 3, 5]

# %%
plt.figure()   # Start a new figure
# Plot y versus x (called "Series 1" in the legend).
#   b = blue.  o = show points.  - = show line.
plt.plot(x, y, 'bo-', label="Series 1")
# Plot y2 versus x (called "Series 2" in the legend).
#   r = red.  x = show points as x'es.   -- = show dashed line
plt.plot(x, y2, 'rx--', label="Series 2")
plt.title("Title of the plot")
plt.xlabel("Label of the x-axis")
plt.ylabel("Label of the y-axis")
# plt.ylim(0, 10)     # Uncomment to set range of y values to show - similar for x.
plt.legend()  # Make a legend, let matplotlib decide where to place it.
plt.show()

# %%
"""
You can save the plot as a figure by pressing the "floppy disk" icon.  Note that in some browsers it does not work, then you can stop the interactive plot by clicking the blue on/off button in the upper right corner, and then save the plot as any other figure in your browser (probably by left-clicking it).

Sometimes, you need larger fonts in a plot that you want to include in a report.  Below is the same plot, but with larger fonts.  We also overrule the placement of the legend.
"""

# %%
plt.figure()
plt.plot(x, y, 'bo-', label="Series 1")
plt.plot(x, y2, 'rx--', label="Series 2")
plt.title("Title of the plot", size=24)
plt.xlabel("Label of the x-axis", size=16)
plt.ylabel("Label of the y-axis", size=16)
# Make tick marks larger, and increase the font.
plt.tick_params(axis='both', which='major', labelsize=16, size=10)
plt.ylim(-0.2, 10.2)     # Set range of y values to show - similar for x.
plt.legend(loc="upper left", fontsize=16)  # Make a legend, specify location.
plt.tight_layout()   # Fixes that otherwise some of the labels are cropped.
plt.show()

# %%
"""
More advanced example with multiple sub-plots.
"""

# %%
fig, axs = plt.subplots(1, 2, sharey=True)
x = np.linspace(0, 2 * np.pi, 100)
axs[0].plot(x, np.cos(x), label='cos')
axs[1].plot(x, np.sin(x), label='sin')
axs[0].legend()
axs[1].legend()
plt.show()

# %%
"""
Plotting a countour
"""

# %%
x = np.linspace(-1, 1, 100)
y = np.linspace(-2, 2, 100)
X, Y = np.meshgrid(x, y)
Z = X**2 + Y**2
N = 15

fig, ax = plt.subplots(1, 1)
ax.contour(X, Y, Z, N)
ax.set_aspect('equal')

# %%
"""
# ASE (atomic simulation environment)
More details can be found here: https://wiki.fysik.dtu.dk/ase/index.html
"""

# %%
"""
## Everything starts with a structure!
In ASE the most important ingredients is the `Atom` and `Atoms`
objects used to setup an atomic structure.
### Setting op a molecule using the `Atoms` object
"""

# %%
from ase import Atoms
d = 1.1
co = Atoms('CO', positions=[[0, 0, 0], [0, 0, d]])

# %%
# lets try and visualize it using the build in viewer in ase
from ase.visualize import view
view(co)

# %%
"""
### Setting up a periodic structure
"""

# %%
d = 2.9
L = 10
wire = Atoms('Au', positions=[[0, L / 2, L / 2]],
             cell=[d, L, L], pbc=[1, 0, 0])
# lets try and repeat it and visualize primitive and repeated
wire10 = wire * (10, 1, 1)
view([wire, wire10])

# %%
"""
## Nitrogen on copper
### Exercise of the relaxation of a molecule on a surface
This section gives a quick (and incomplete) overview of what ASE can do.

We will calculate the adsorption energy of a nitrogen molecule on a copper
surface. This is done by calculating the total energy for the isolated slab
and for the isolated molecule. The adsorbate is then added to the slab and
relaxed, and the total energy for this composite system is calculated. The
adsorption energy is obtained as the sum of the isolated energies minus the
energy of the composite system.

You can read more about the optimizers in ASE here:
https://wiki.fysik.dtu.dk/ase/ase/optimize.html

#### 1. Try to go through the script so you understand what is going on
#### 2. Calculate the adsorption energy of N2 on a 4x4x2 fcc111 slab (result= 0.324 eV)
#### 3. Try a couple of different optimizers and see which one is the fastest
"""

# %%
from ase import Atoms
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton, BFGS, FIRE
from ase.build import fcc111, add_adsorbate
from ase.visualize import view

h = 2.85
d = 1.10

# Set up a slab:
slab = fcc111('Cu', size=(4, 4, 2), vacuum=10.0)
slab.calc = EMT()
e_slab = slab.get_potential_energy()
# setup a molecule
molecule = Atoms('2N', positions=[(0., 0., 0.), (0., 0., d)])
molecule.calc = EMT()
e_N2 = molecule.get_potential_energy()

add_adsorbate(slab, molecule, h, 'ontop')
constraint = FixAtoms(mask=[a.symbol != 'N' for a in slab])
slab.set_constraint(constraint)
dyn = QuasiNewton(slab, trajectory='N2Cu.traj')
dyn.run(fmax=0.025)

print('Adsorption energy:', e_slab + e_N2 - slab.get_potential_energy())

# %%
# Visualize the trajectory
# magic: !ase gui N2Cu.traj

# %%
"""
## Band structure
#### Using ASE to setup band structures for Al using a Freelectron model and DFT
#### 1. What is the crystal structure of Al?
#### 2. Try and look up the recommeded Brillouin zone path for crystal structure [here](https://wiki.fysik.dtu.dk/ase/ase/dft/kpoints.html)
#### 3. Can you figure out what the `nbands=-10` and `convergence={'bands': -5}` parameters means in the GPAW DFT input below ? (Hint try and look at the output file `Al.txt`)
"""

# %%
# simple free electron calculator
from ase.build import bulk
from ase.calculators.test import FreeElectrons

a = bulk('Al')
kpts = {'path': 'GXWLGK', 'npoints': 100}

# Simple FreeElectron model calculator
a.calc = FreeElectrons(nvalence=3,
                       kpts=kpts)
a.get_potential_energy()
bs = a.calc.band_structure()
bs.plot(emax=10, filename='al-free-electron.png')

# %%
# setup a DFT calculation with GPAW and repeat
from gpaw import GPAW, PW
# calc the self-consistent electron density
a.calc = GPAW(kpts=(3, 3, 3), mode=PW(200), txt='Al.txt')
a.get_potential_energy()
# band-structure calculation for a fixed density
calc = a.calc.fixed_density(
    kpts=kpts,
    symmetry='off',
    nbands=-10,
    convergence={'bands': -5})

# %%
bs = calc.band_structure()
bs.plot(emax=10, filename='al-dft.png')



import sys

import ase.db
from ase.optimize.test.test import (test_optimizer, all_optimizers,
                                    get_optimizer)
from gpaw import GPAW


db1 = ase.db.connect('systems.db')
db = ase.db.connect('results-lcao.db')


def lcao(txt):
    return GPAW(mode='lcao',
                basis='dzp',
                kpts={'density': 2.0},
                txt=txt)


systems = [(row.name, row.toatoms()) for row in db1.select()]

if len(sys.argv) > 1:
    optimizers = sys.argv[1:]
else:
    optimizers = all_optimizers

for opt in optimizers:
    optimizer = get_optimizer(opt)
    test_optimizer(systems, optimizer, lcao, 'lcao-', db)



from ase.build import molecule
from ase.optimize import QuasiNewton
from gpaw import GPAW

atoms = molecule('CH3CH2OH', vacuum=4.0)
atoms.rattle(stdev=0.1)  # displace positions randomly a bit

calc = GPAW(mode='lcao',
            basis='dzp',
            nbands='110%',
            parallel=dict(band=2,  # band parallelization
                          augment_grids=True,  # use all cores for XC/Poisson
                          sl_auto=True,  # enable ScaLAPACK parallelization
                          use_elpa=True))  # enable Elpa eigensolver
atoms.calc = calc

opt = QuasiNewton(atoms, trajectory='opt.traj')
opt.run(fmax=0.05)



from ase.build import bulk
from gpaw import GPAW
from gpaw.elph.electronphonon import ElectronPhononCoupling

atoms = bulk('C', 'diamond', a=3.567)
calc = GPAW(mode='lcao', basis='dzp',
            kpts=(3, 3, 3), xc='PBE',
            symmetry={'point_group': False},
            txt='elph.txt')

# Displacements for potential
elph = ElectronPhononCoupling(atoms, calc, supercell=(2, 2, 2),
                              calculate_forces=False)
elph.run()



from ase.dft import monkhorst_pack
import numpy as np
from ase import Atoms
from gpaw import GPAW, FermiDirac
from gpaw import PW
from math import sqrt

# Part 1: Ground state calculation

NBN = 7
NGr = 7
a = 2.5
c = 3.22

GR = Atoms(symbols='C2',
           positions=[(0.5 * a, -sqrt(3) / 6 * a, 0.0),
                      (0.5 * a, +sqrt(3) / 6 * a, 0.0)],
           cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                 (0.5 * a, +0.5 * 3**0.5 * a, 0), (0.0, 0.0, c * 2.0)])
GR.set_pbc((True, True, True))

GR2 = GR.copy()
cell = GR2.get_cell()
uv = cell[0] - cell[1]
uv = uv / np.sqrt(np.sum(uv**2.0))
dist = np.array([0.5 * a, -sqrt(3) / 6 * a]) - np.array(
    [0.5 * a, +sqrt(3) / 6 * a])
dist = np.sqrt(np.sum(dist**2.0))
GR2.translate(uv * dist)

BN = Atoms(symbols='BN',
           positions=[(0.5 * a, -sqrt(3) / 6 * a, 0.0),
                      (0.5 * a, +sqrt(3) / 6 * a, 0.0)],
           cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                 (0.5 * a, +0.5 * 3**0.5 * a, 0), (0.0, 0.0, c * 2.0)])
BN.set_pbc((True, True, True))

NB = Atoms(symbols='NB',
           positions=[(0.5 * a, -sqrt(3) / 6 * a, 0.0),
                      (0.5 * a, +sqrt(3) / 6 * a, 0.0)],
           cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                 (0.5 * a, +0.5 * 3**0.5 * a, 0), (0.0, 0.0, c * 2.0)])
NB.set_pbc((True, True, True))

GR2.translate([0, 0, c])
NB.translate([0, 0, (NGr + 1.0 * (1 - NGr % 2)) * c] + uv * dist * (NGr % 2))
BN.translate([0, 0, (NGr + 1.0 * (NGr % 2)) * c] + uv * dist * (NGr % 2))

GRBN = (GR * (1, 1, (NGr % 2 + NGr // 2)) + GR2 * (1, 1, (NGr // 2)) + NB *
        (1, 1, (NBN // 2 + (NBN % 2) * (NGr % 2))) + BN *
        (1, 1, (NBN // 2 + (NBN % 2) * (1 - NGr % 2))))
BNNB = BN + NB
Graphite = GR + GR2

Graphite.set_pbc((True, True, True))
old_cell = GR.get_cell()
old_cell[2, 2] = 2 * c
Graphite.set_cell(old_cell)

BNNB.set_pbc((True, True, True))
old_cell = BN.get_cell()
old_cell[2, 2] = 2 * c
BNNB.set_cell(old_cell)
BNNB.center()

GRBN.set_pbc((True, True, True))
old_cell = BN.get_cell()
old_cell[2, 2] = (NGr + NBN) * c
GRBN.set_cell(old_cell)

atoms = GRBN

calc = GPAW(
    h=0.18,
    mode=PW(600),
    kpts=monkhorst_pack((29, 29, 1)) + np.array([0., 0., 0.]),
    xc='PBE',
    occupations=FermiDirac(0.01),
    parallel={'band': 1},
)

atoms.calc = calc

ncpus = 16



import numpy as np
from gpaw.blacs import BlacsDescriptor, BlacsGrid, Redistributor


class Blocks1D:
    def __init__(self, blockcomm, N):
        self.blockcomm = blockcomm
        self.N = N  # Global number of points

        self.blocksize = (N + blockcomm.size - 1) // blockcomm.size
        self.a = min(blockcomm.rank * self.blocksize, N)
        self.b = min(self.a + self.blocksize, N)
        self.nlocal = self.b - self.a

        self.myslice = slice(self.a, self.b)

    def collect(self, array_w):
        b_w = np.zeros(self.blocksize, array_w.dtype)
        b_w[:self.nlocal] = array_w
        A_w = np.empty(self.blockcomm.size * self.blocksize, array_w.dtype)
        self.blockcomm.all_gather(b_w, A_w)
        return A_w[:self.N]


def block_partition(comm, nblocks):
    assert comm.size % nblocks == 0, comm.size
    rank1 = comm.rank // nblocks * nblocks
    rank2 = rank1 + nblocks
    blockcomm = comm.new_communicator(range(rank1, rank2))
    ranks = range(comm.rank % nblocks, comm.size, nblocks)
    if nblocks == 1:
        assert len(ranks) == comm.size
        intrablockcomm = comm
    else:
        intrablockcomm = comm.new_communicator(ranks)
    assert blockcomm.size * intrablockcomm.size == comm.size
    return blockcomm, intrablockcomm


class PlaneWaveBlockDistributor:
    """Functionality to shuffle block distribution of pair functions
    in the plane wave basis."""

    def __init__(self, world, blockcomm, intrablockcomm):
        self.world = world
        self.blockcomm = blockcomm
        self.intrablockcomm = intrablockcomm

    def redistribute(self, in_wGG, nw):
        """Redistribute array.

        Switch between two kinds of parallel distributions:

        1) parallel over G-vectors (second dimension of in_wGG)
        2) parallel over frequency (first dimension of in_wGG)

        Returns new array using the memory in the 1-d array out_x.
        """

        comm = self.blockcomm

        if comm.size == 1:
            return in_wGG

        mynw = (nw + comm.size - 1) // comm.size
        nG = in_wGG.shape[2]
        mynG = (nG + comm.size - 1) // comm.size

        bg1 = BlacsGrid(comm, comm.size, 1)
        bg2 = BlacsGrid(comm, 1, comm.size)
        md1 = BlacsDescriptor(bg1, nw, nG**2, mynw, nG**2)
        md2 = BlacsDescriptor(bg2, nw, nG**2, nw, mynG * nG)

        if len(in_wGG) == nw:
            mdin = md2
            mdout = md1
        else:
            mdin = md1
            mdout = md2

        r = Redistributor(comm, mdin, mdout)

        # mdout.shape[1] is always divisible by nG because
        # every block starts at a multiple of nG, and the last block
        # ends at nG² which of course also is divisible.  Nevertheless:
        assert mdout.shape[1] % nG == 0
        # (If it were not divisible, we would "lose" some numbers and the
        #  redistribution would be corrupted.)

        outshape = (mdout.shape[0], mdout.shape[1] // nG, nG)
        out_wGG = np.empty(outshape, complex)

        inbuf = in_wGG.reshape(mdin.shape)
        outbuf = out_wGG.reshape(mdout.shape)

        r.redistribute(inbuf, outbuf)

        return out_wGG

    def distribute_frequencies(self, in_wGG, nw):
        """Distribute frequencies to all cores."""

        world = self.world
        comm = self.blockcomm

        if world.size == 1:
            return in_wGG

        mynw = (nw + world.size - 1) // world.size
        nG = in_wGG.shape[2]
        mynG = (nG + comm.size - 1) // comm.size

        wa = min(world.rank * mynw, nw)
        wb = min(wa + mynw, nw)

        if self.blockcomm.size == 1:
            return in_wGG[wa:wb].copy()

        if self.intrablockcomm.rank == 0:
            bg1 = BlacsGrid(comm, 1, comm.size)
            in_wGG = in_wGG.reshape((nw, -1))
        else:
            bg1 = BlacsGrid(None, 1, 1)
            # bg1 = DryRunBlacsGrid(mpi.serial_comm, 1, 1)
            in_wGG = np.zeros((0, 0), complex)
        md1 = BlacsDescriptor(bg1, nw, nG**2, nw, mynG * nG)

        bg2 = BlacsGrid(world, world.size, 1)
        md2 = BlacsDescriptor(bg2, nw, nG**2, mynw, nG**2)

        r = Redistributor(world, md1, md2)
        shape = (wb - wa, nG, nG)
        out_wGG = np.empty(shape, complex)
        r.redistribute(in_wGG, out_wGG.reshape((wb - wa, nG**2)))

        return out_wGG



# web-page: Fe_magnons_vs_rc.png
"""Plot the magnon energy as a function of the cutoff radius rc for all
high-symmetry points of Fe (bcc)"""

# General modules
import numpy as np
import matplotlib.pyplot as plt

# Script modules
from gpaw import GPAW
from gpaw.response.heisenberg import calculate_single_site_magnon_energies

# ---------- Inputs ---------- #

# Ground state
gpw = 'Fe_all.gpw'

# High-symmetry points
sp_p = ['N', 'P', 'H']
spq_pc = [np.array(qc) for qc in
          [[0., 0., 0.5], [0.25, 0.25, 0.25], [0.5, -0.5, 0.5]]]

# Load MFT data
q_qc = np.load('Fe_q_qc.npy')
rc_r = np.load('Fe_rc_r.npy')
J_qr = np.load('Fe_J_qr.npy')

# Labels and limits
rlabel = r'$r_{\mathrm{c}}\: [\mathrm{\AA}]$'
mwlabel = r'$\hbar\omega$ [meV]'
rlim = (0.4, 1.85)
mwlim = (200., 600.)

filename = 'Fe_magnons_vs_rc.png'

# ---------- Script ---------- #

# Extract the magnetization of the unit cell
calc = GPAW(gpw, txt=None)
muc = calc.get_magnetic_moment()

# Calculate the magnon energies
E_qr = calculate_single_site_magnon_energies(J_qr, q_qc, muc)

# Plot the magnon energies at the high-symmetry points as a function of rc
for sp, spq_c in zip(sp_p, spq_pc):
    q = 0
    while not np.allclose(q_qc[q], spq_c):
        q += 1
        if q == len(q_qc):
            raise ValueError
    E_r = E_qr[q] * 1.e3  # eV -> meV
    plt.plot(rc_r, E_r, '-x', label=sp)

plt.xlabel(rlabel)
plt.ylabel(mwlabel)
plt.xlim(rlim)
plt.ylim(mwlim)

plt.legend()

plt.savefig(filename, format=filename.split('.')[-1],
            bbox_inches='tight')



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw import PW
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy

from ase.parallel import paropen

resultfile = paropen('si.pbe+exx.results.txt', 'a')

# Plane wave cutoff
pwcutoff = 400.0

# NxNxN k-point sampling, gamma-centred grid
k = 6

# Si lattice constant
alat = 5.421

# Do the bulk calculation

bulk_crystal = bulk('Si', 'diamond', a=alat)
bulk_calc = GPAW(mode=PW(pwcutoff),
                 kpts={'size': (k, k, k), 'gamma': True},
                 xc='PBE',
                 occupations=FermiDirac(0.01),
                 txt='si.pbe+exx.pbe_output.txt',
                 parallel={'band': 1}
                 )

bulk_crystal.calc = bulk_calc
e0_bulk_pbe = bulk_crystal.get_potential_energy()

#  Write to file
bulk_calc.write('bulk.gpw', mode='all')

# Now the exact exchange
e0_bulk_exx = nsc_energy('bulk.gpw', 'EXX')

s = str(alat)
s += ' '
s += str(k)
s += ' '
s += str(pwcutoff)
s += ' '
s += str(e0_bulk_pbe)
s += ' '
s += str(e0_bulk_exx)
s += '\n'
resultfile.write(s)



"""Test the calculation of the excitation energy of Na2 by RSF and IVOs."""
from ase import Atoms
from ase.units import Hartree
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.occupations import FermiDirac
from gpaw.test import equal, gen
from gpaw.eigensolvers import RMMDIIS
from gpaw.cluster import Cluster
from gpaw.lrtddft import LrTDDFT
import _gpaw


def test_rsf_yukawa_rsf_ivo_sing_mg(in_tmp_dir, add_cwd_to_setup_paths):
    libxc_version = getattr(_gpaw, 'libxc_version', '2.x.y')
    if int(libxc_version.split('.')[0]) < 3:
        from unittest import SkipTest
        raise SkipTest

    h = 0.35  # Gridspacing
    e_singlet = 4.61
    e_singlet_lr = 5.54

    gen('Mg', xcname='PBE', scalarrel=True, exx=True, yukawa_gamma=0.38)

    c = {'energy': 0.05, 'eigenstates': 3, 'density': 3}
    na2 = Cluster(Atoms('Mg', positions=[[0, 0, 0]]))
    na2.minimal_box(2.5, h=h)
    calc = GPAW(txt='mg_ivo.txt', xc='LCY-PBE:omega=0.38:excitation=singlet',
                eigensolver=RMMDIIS(), h=h, occupations=FermiDirac(width=0.0),
                spinpol=False, convergence=c)
    na2.calc = calc
    na2.get_potential_energy()
    (eps_homo, eps_lumo) = calc.get_homo_lumo()
    e_ex = eps_lumo - eps_homo
    equal(e_singlet, e_ex, 0.15)
    calc.write('mg.gpw')

    c2 = GPAW('mg.gpw')
    ihomo = int(c2.get_occupation_numbers().sum() / 2 + 0.5) - 1
    assert c2.hamiltonian.xc.excitation == 'singlet'
    lr = LrTDDFT(calc, txt='LCY_TDDFT_Mg.log',
                 restrict={'istart': ihomo, 'jend': ihomo + 1}, nspins=2)
    lr.write('LCY_TDDFT_Mg.ex.gz')
    if world.rank == 0:
        lr2 = LrTDDFT.read('LCY_TDDFT_Mg.ex.gz')
        lr2.diagonalize()
        ex_lr = lr2[1].get_energy() * Hartree
        equal(e_singlet_lr, ex_lr, 0.15)



# creates: H.rst, H.default.png
# ... and all the rest.
import json
import sys

import matplotlib.pyplot as plt
from ase.data import atomic_numbers, atomic_names
from ase.units import Hartree
from ase.utils import plural

from gpaw.atom.check import cutoffs, all_names


with open('datasets.json') as fd:
    data = json.load(fd)


def rst(symbol):
    Z = atomic_numbers[symbol]
    name = atomic_names[Z]

    rst = """\
.. Computer generated reST (make_setup_pages.py)
.. index:: {name}
.. _{name}:

================
{name}
================

Datasets:

.. csv-table::
    :header: name, valence electrons, frozen core electrons

{table}"""

    table = ''
    for e, nlfer, energies in data[symbol]:
        nv, txt = rst1(symbol + '.' + e, nlfer, energies)
        if e != 'default':
            e = f"``'{e}'``"
        table += f'    {e},{nv},{Z - nv}\n'
        rst += txt

    with open(symbol + '.rst', 'w') as fd:
        fd.write(rst.format(table=table, name=name))


def rst1(dataset, nlfer, energies):
    table1 = ''
    nv = 0
    for n, l, f, e, rcut in nlfer:
        n, l, f = (int(x) for x in [n, l, f])
        if n == -1:
            n = ''
        table1 += f"    {n}{'spdf'[l]},{f},{e * Hartree:.3f},"
        if rcut:
            table1 += f'{rcut:.2f}'
            nv += f
        table1 += '\n'

    rst = """

{electrons}
====================

Radial cutoffs and eigenvalues:

.. csv-table::
    :header: id, occ, eig [eV], cutoff [Bohr]

{table1}

The figure shows convergence of the absolute energy (red line)
and atomization energy (green line) of a {symbol} dimer relative
to completely converged numbers (plane-wave calculation at 1500 eV).
Also shown are finite-difference and LCAO (dzp) calculations at gridspacings
0.143 Å, 0.167 Å and 0.200 Å.

.. image:: {dataset}.png

Egg-box errors in finite-difference mode:

.. csv-table::
    :header: grid-spacing [Å], energy error [eV]

{table2}"""

    epw, depw, efd, defd, elcao, delcao, deegg = energies

    table2 = ''
    for h, e in zip([0.16, 0.18, 0.2], deegg):
        table2 += f'    {h:.2f},{e:.4f}\n'

    fig = plt.figure(figsize=(8, 5))

    ax1 = plt.subplot(121)
    ax1.semilogy(cutoffs[:-1], epw[:-1], 'C0-',
                 label='PW, absolute')
    ax1.semilogy(cutoffs[:-1], depw[:-1], 'C1--',
                 label='PW, atomization')
    plt.xticks([200, 400, 600, 800], fontsize=15)
    plt.yticks(fontsize=15)
    plt.xlabel('Planewave cutoff [eV]', fontsize=15)
    plt.ylabel('Error [eV/atom]', fontsize=15)
    plt.legend(loc='best')

    ax2 = plt.subplot(122, sharey=ax1)
    h = [4.0 / g for g in [20, 24, 28]]
    ax2.semilogy(h, efd, 'C0s-', label='FD, absolute')
    ax2.semilogy(h, defd, 'C1s--', label='FD, atomization')
    ax2.semilogy(h, elcao, 'C0o-', label='LCAO, absolute')
    ax2.semilogy(h, delcao, 'C1o--', label='LCAO, atomization')
    plt.xticks([0.16, 0.18, 0.2], fontsize=15)
    plt.xlim(0.14, 0.2)
    plt.xlabel(u'grid-spacing [Å]', fontsize=15)
    plt.legend(loc='best')
    plt.setp(ax2.get_yticklabels(), visible=False)

    plt.tight_layout()
    plt.subplots_adjust(wspace=0)
    plt.savefig(dataset + '.png')
    plt.close(fig)

    return nv, rst.format(electrons=plural(nv, 'valence electron'),
                          table1=table1, table2=table2, symbol=symbol,
                          dataset=dataset)


for symbol in all_names:
    if '.' not in symbol:
        print(symbol, end='')
        sys.stdout.flush()
        rst(symbol)
print()



import numpy as np


"""This module exists to reduce code duplication.

The costruct_symmetry_operators method was duplicated on two classes.
Now it only exists in one place, at the cost of a nasty fake "self".
"""


def construct_symmetry_operators(gs, K, k_c=None, *, apply_strange_shift):
    """Construct symmetry operators for wave function and PAW projections.

    We want to transform a k-point in the irreducible part of the BZ to
    the corresponding k-point with index K.

    Returns U_cc, T, a_a, U_aii, shift_c and time_reversal, where:

    * U_cc is a rotation matrix.
    * T() is a function that transforms the periodic part of the wave
      function.
    * a_a is a list of symmetry related atom indices
    * U_aii is a list of rotation matrices for the PAW projections
    * shift_c is three integers: see code below.
    * time_reversal is a flag - if True, projections should be complex
      conjugated.

    See the get_k_point() method for how to use these tuples.
    """

    R_asii = [setup.R_sii for setup in gs.setups]
    return _construct_symmetry_operators(
        gs, K, k_c=k_c,
        apply_strange_shift=apply_strange_shift,
        R_asii=R_asii)


def _construct_symmetry_operators(gs, K, k_c=None, *, apply_strange_shift,
                                  R_asii):
    kd = gs.kd

    s = kd.sym_k[K]
    U_cc = kd.symmetry.op_scc[s]
    time_reversal = kd.time_reversal_k[K]
    ik = kd.bz2ibz_k[K]
    if k_c is None:
        k_c = kd.bzk_kc[K]
    ik_c = kd.ibzk_kc[ik]

    sign = 1 - 2 * time_reversal
    shift_c = np.dot(U_cc, ik_c) - k_c * sign

    assert np.allclose(shift_c.round(), shift_c)

    shift_c = shift_c.round().astype(int)

    if (U_cc == np.eye(3)).all():
        def T(f_R):
            return f_R
    else:
        N_c = gs.gd.N_c
        i_cr = np.dot(U_cc.T, np.indices(N_c).reshape((3, -1)))
        i = np.ravel_multi_index(i_cr, N_c, 'wrap')

        def T(f_R):
            return f_R.ravel()[i].reshape(N_c)

    if time_reversal:
        T0 = T

        def T(f_R):
            return T0(f_R).conj()

        shift_c *= -1

    a_a = []
    U_aii = []
    for a, R_sii in enumerate(R_asii):
        b = kd.symmetry.a_sa[s, a]
        S_c = np.dot(gs.spos_ac[a], U_cc) - gs.spos_ac[b]
        x = np.exp(2j * np.pi * np.dot(ik_c, S_c))
        U_ii = R_sii[s].T * x
        a_a.append(b)
        U_aii.append(U_ii)

    if apply_strange_shift:
        shift0_c = (kd.bzk_kc[K] - k_c).round().astype(int)
        shift_c += -shift0_c

    return U_cc, T, a_a, U_aii, shift_c, time_reversal



from ase import Atoms
from gpaw import GPAW
from gpaw.lrtddft import LrTDDFT

# Na2 cluster
atoms = Atoms(symbols='Na2',
              positions=[(0, 0, 0), (3.0, 0, 0)],
              pbc=False)
atoms.center(vacuum=6.0)

# Standard ground state calculation with empty states
calc = GPAW(nbands=100, h=0.4, setups={'Na': '1'})
atoms.calc = calc
energy = atoms.get_potential_energy()

calc = calc.fixed_density(
    convergence={'bands': 90})
calc.write('na2_gs_casida.gpw', mode='all')

# Standard Casida calculation
calc = GPAW('na2_gs_casida.gpw')
istart = 0
jend = 90
lr = LrTDDFT(calc, xc='LDA', restrict={'istart': istart, 'jend': jend})
lr.diagonalize()
lr.write('na2_lr.dat.gz')



""" TODO:

1. we should find a good way in which to store files elsewhere than static

2. currently the files that are not generated by weekly tests are copied
   from srcpath. This needs to be documented.

Make sure that downloaded files are copied to build dir on build
This must (probably) be done *after* compilation because otherwise dirs
may not exist.

"""
try:
    from urllib2 import urlopen, HTTPError
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError
    import ssl
import os


srcpath = 'https://wiki.fysik.dtu.dk/gpaw-files'
agtspath = 'http://wiki.fysik.dtu.dk'


def get(path, names, target=None, source=None):
    """Get files from web-server.

    Returns True if something new was fetched."""

    if target is None:
        target = path
    if source is None:
        source = srcpath
    got_something = False
    # We get images etc from a web server with a self-signed certificate
    # That cause trouble on some machines.
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    for name in names:
        src = os.path.join(source, path, name)
        dst = os.path.join(target, name)

        if not os.path.isfile(dst):
            print(dst, end=' ')
            try:
                data = urlopen(src, context=ctx).read()
                with open(dst, 'wb') as sink:
                    sink.write(data)
                print('OK')
                got_something = True
            except HTTPError:
                print('HTTP Error!')
    return got_something


literature = """
askhl_10302_report.pdf  mortensen_gpaw-dev.pdf      rostgaard_master.pdf
askhl_master.pdf        mortensen_mini2003talk.pdf
marco_master.pdf        mortensen_paw.pdf           ss14.pdf
""".split()
get('doc/literature', literature, 'documentation')

# Note: bz-all.png is used both in an exercise and a tutorial.  Therefore
# we put it in the common dir so far, rather than any of the two places
get('.', ['bz-all.png'], 'static')

# These files have different destinations after webpage refactor then they
# used to have. Might need to keep track of this.
get('exercises/wavefunctions', ['co_bonding.jpg'],
    target='tutorialsexercises/wavefunctions/wavefunctions')
get('exercises/lrtddft', ['spectrum.png'],
    target='tutorialsexercises/opticalresponse/lrtddft')
get('tutorials/wannier90', ['GaAs.png', 'Cu.png', 'Fe.png'],
    target='tutorialsexercises/wavefunctions/wannier90')
get('tutorials/xas', ['h2o_xas_3.png', 'h2o_xas_4.png'],
    target='tutorialsexercises/opticalresponse/xas')
get('tutorials/xas',
    ['xas_illustration.png'], target='documentation/xas')

# This files is not used anymore?
# get('tutorialsexercises/opticalresponse/xas', ['xas_h2o_convergence.png'])
# ----

get('documentation/xc', 'g2test_pbe0.png  g2test_pbe.png  results.png'.split())
get('performance', 'dacapoperf.png  goldwire.png  gridperf.png'.split(),
    'static')

get('bgp', ['bgp_mapping_intranode.png',
            'bgp_mapping1.png',
            'bgp_mapping2.png'], 'platforms/BGP')

# workshop 2013 and 2016 photos:
get('workshop13', ['workshop13_01_33-1.jpg'], 'static')
get('workshop16', ['gpaw2016-photo.jpg'], 'static')


# files from https://wiki.fysik.dtu.dk/gpaw-files/things/

# Warning: for the moment dcdft runs are not run (files are static)!
dcdft_pbe_aims_stuff = """
dcdft_aims.tight.01.16.db.csv
dcdft_aims.tight.01.16.db_raw.csv
dcdft_aims.tight.01.16.db_Delta.txt
""".split()

get('things', dcdft_pbe_aims_stuff, target='setups')

# Warning: for the moment dcdft runs are not run (files are static)!
dcdft_pbe_gpaw_pw_stuff = """
dcdft_pbe_gpaw_pw.csv
dcdft_pbe_gpaw_pw_raw.csv
dcdft_pbe_gpaw_pw_Delta.txt
""".split()

get('things', dcdft_pbe_gpaw_pw_stuff, target='setups')

# Warning: for the moment dcdft runs are not run (files are static)!
dcdft_pbe_jacapo_stuff = """
dcdft_pbe_jacapo.csv
dcdft_pbe_jacapo_raw.csv
dcdft_pbe_jacapo_Delta.txt
""".split()

get('things', dcdft_pbe_jacapo_stuff, target='setups')

# Warning: for the moment dcdft runs are not run (files are static)!
dcdft_pbe_abinit_fhi_stuff = """
dcdft_pbe_abinit_fhi.csv
dcdft_pbe_abinit_fhi_raw.csv
dcdft_pbe_abinit_fhi_Delta.txt
""".split()

get('things', dcdft_pbe_abinit_fhi_stuff, target='setups')

g2_1_stuff = """
pbe_gpaw_nrel_ea_vs.csv pbe_gpaw_nrel_ea_vs.png
pbe_gpaw_nrel_opt_ea_vs.csv pbe_gpaw_nrel_opt_distance_vs.csv
pbe_nwchem_def2_qzvppd_opt_ea_vs.csv pbe_nwchem_def2_qzvppd_opt_distance_vs.csv
""".split()

get('things', g2_1_stuff, target='setups')

get('things', ['datasets.json'], 'setups')

# Carlsberg foundation figure:
get('.', ['carlsberg.png'])

get('static', ['NOMAD_Logo_supported_by.png'])

# Summer school 2022
get('summerschool2018',
    ['CreateTunnelWin.png', 'JupyterRunningMac.png', 'JupyterRunningWin.png',
     'Logged_in_Mac.png', 'Logged_in_Win.png', 'Moba_ssh.png',
     'UseTunnelWin.png'],
    target='summerschools/summerschool22')
get('summerschool2018',
    ['organometal.master.db'],
    target='summerschools/summerschool22/machinelearning')
get('summerschool2018',
    ['C144Li18.png', 'C64.png', 'final.png', 'initial.png',
     'Li2.png', 'lifepo4_wo_li.traj', 'NEB_init.traj'],
    target='summerschools/summerschool22/batteries')
get('summerschool2022',
    ['Intro_projects_CAMD2022.pdf'],
    target='summerschools/summerschool22')


def setup(app):
    # Get png and csv files and other stuff from the AGTS scripts that run
    # every weekend:
    from gpaw.doctools.agts_crontab import find_created_files

    for path in find_created_files():
        # the files are saved by the weekly tests under agtspath/agts-files
        # now we are copying them back to their original run directories
        if path.is_file():
            continue
        print(path, 'copied from', agtspath)
        get('agts-files', [path.name], str(path.parent), source=agtspath)



import numpy as np
from typing import TypeVar, Any
def adjust_positions() -> None: ...
def adjust_momenta() -> None: ...
def calculate_forces_H2O() -> None: ...
def localize(Z_nnc: Any, U_nn: Any) -> float: ...
def spherical_harmonics() -> None: ...
def hartree(l: int, nrdr: np.ndarray, r: np.ndarray, vr: np.ndarray) -> None: ...
def get_num_threads() -> int: ...
def pack(A: np.ndarray) -> np.ndarray: ...
T = TypeVar('T', float, complex)
def mmm(alpha: T,
        a: np.ndarray,
        opa: str,
        b: np.ndarray,
        opb: str,
        beta: T,
        c: np.ndarray) -> None: ...
def gemm(alpha, a, b, beta, c, transa='n') -> None: ...
def rk(alpha, a, beta, c, trans='c') -> None: ...
def r2k(alpha, a, b, beta, c, trans='c') -> None: ...
def gemmdot() -> None: ...
class Communicator:
    rank: int
    size: int
def add_to_density(f: float, psit: np.ndarray, density: np.ndarray) -> None: ...
def pw_insert(coef_G: np.ndarray, Q_G: np.ndarray, s: float, array_Q: np.ndarray) -> None: ...
def pblas_tran(N: int, M: int,
               alpha: float, a_MN: np.ndarray,
               beta:float, c_NM: np.ndarray,
               desca: np.ndarray, descc: np.ndarray,
               conj: bool) -> None: ...
def scalapack_set(a: np.ndarray, desc: np.ndarray, alpha: float, beta: float,
                  uplo: str, n: int, m: int, ja: int, ia: int) -> None: ...
def scalapack_diagonalize_dc(a: np.ndarray, desc: np.ndarray, uplo: str,
                             data: np.ndarray, eps: np.ndarray) -> int: ...



from gpaw.lcao.generate_ngto_augmented import \
    create_GTO_dictionary as GTO, generate_nao_ngto_basis

gtos_atom = {'H': [GTO('s', 0.02974),
                   GTO('p', 0.14100)],
             'C': [GTO('s', 0.04690),
                   GTO('p', 0.04041),
                   GTO('d', 0.15100)],
             'O': [GTO('s', 0.07896),
                   GTO('p', 0.06856),
                   GTO('d', 0.33200)],
             }

for atom in ['H', 'C', 'O']:
    generate_nao_ngto_basis(atom, xc='PBE', name='aug',
                            nao='dzp', gtos=gtos_atom[atom])



from gpaw import GPAW
from ase.build import fcc100


def energy(N, k, a=4.05):
    fcc = fcc100('Al', (1, 1, N), a=a, vacuum=7.5)
    fcc.center(axis=2)
    calc = GPAW(nbands=N * 3,
                kpts=(k, k, 1),
                h=0.25,
                txt='slab-%d.txt' % N)
    fcc.calc = calc
    e = fcc.get_potential_energy()
    calc.write('slab-%d.gpw' % N)
    return e


N = ...
k = ...
e = energy(N, k)



# web-page: mag1d.png, mag2d.png
from gpaw.new.ase_interface import GPAW
import matplotlib.pyplot as plt
import numpy as np

calc = GPAW('VCl2_gs.gpw')
dens = calc.calculation.densities()
grid_spacing = calc.atoms.cell[2, 2] / 200
nt = dens.pseudo_densities(grid_spacing)
n = dens.all_electron_densities(grid_spacing=grid_spacing)

i = nt.desc.size[2] // 2
x, y = n.desc.xyz()[:, :, i, :2].transpose((2, 0, 1))
uv = n.data[1:3, :, :, i]
m = (uv**2).sum(0)**0.5
u, v = uv / m
fig, ax = plt.subplots()
ct = ax.contourf(x, y, m)
cbar = fig.colorbar(ct)
cbar.ax.set_ylabel('magnetization [Å$^{-3}$]')
ax.quiver(*(a[::3, ::3] for a in [x, y, u, v]))
ax.axis('equal')
ax.set_xlabel('x [Å]')
ax.set_ylabel('y [Å]')
fig.savefig('mag2d.png')

fig, ax = plt.subplots()
x, y = n.xy(1, ..., 0, i)
x, yt = nt.xy(1, ..., 0, i)
j = len(x) // 2
L = calc.atoms.cell[0, 0]
x = np.concatenate((x[j:] - L, x[:j]))
y = np.concatenate((y[j:], y[:j]))
yt = np.concatenate((yt[j:], yt[:j]))
ax.plot(x, y, label='all-electron')
ax.plot(x, yt, label='pseudo')
ax.legend()
ax.set_xlabel('x [Å]')
ax.set_ylabel('magnetization [Å$^{-3}$]')
fig.savefig('mag1d.png')



def workflow():
    from myqueue.workflow import run
    with run(script='si.py', cores=8, tmax='15m'):
        run(script='wannier-si.py')
        run(script='../../electronic/dos/dos.py', args=['si.gpw'])
    with run(script='benzene.py', cores=8, tmax='15m'):
        run(script='wannier-benzene.py')



from gpaw import GPAW

G = [0.0, 0.0, 0.0]
L = [0.5, 0.0, 0.0]
F = [0.5, 0.5, 0.0]
Z = [0.5, 0.5, 0.5]

calc = GPAW('gs_Bi2Se3.gpw').fixed_density(
    kpts=[G, L, F, Z],
    # symmetry='off',
    txt='high_sym.txt')
calc.diagonalize_full_hamiltonian(nbands=50)

calc.write('high_sym.gpw', mode='all')



import _gpaw
import numpy as np
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.core.plane_waves import PlaneWaveExpansions


class PWHamiltonian(Hamiltonian):
    def apply(self,
              vt_sR: UniformGridFunctions,
              psit_nG: PlaneWaveExpansions,
              out: PlaneWaveExpansions,
              spin: int):
        out_nG = out
        vt_R = vt_sR.data[spin]
        np.multiply(psit_nG.desc.ekin_G, psit_nG.data, out_nG.data)
        grid = vt_sR.desc
        if psit_nG.desc.dtype == complex:
            grid = grid.new(dtype=complex)
        f_R = grid.empty()
        for p_G, o_G in zip(psit_nG, out_nG):
            f_R = p_G.ifft(out=f_R)
            f_R.data *= vt_R
            o_G.data += f_R.fft(pw=p_G.desc).data
        return out_nG

    def create_preconditioner(self, blocksize):
        return precondition


def precondition(psit, residuals, out):
    G2 = psit.desc.ekin_G * 2
    for r, o, ekin in zip(residuals.data, out.data, psit.norm2('kinetic')):
        _gpaw.pw_precond(G2, r, ekin, o)


def spinor_precondition(psit_nsG, residual_nsG, out):
    G2_G = psit_nsG.desc.ekin_G * 2
    for r_sG, o_sG, ekin in zip(residual_nsG.data,
                                out.data,
                                psit_nsG.norm2('kinetic').sum(1)):
        for r_G, o_G in zip(r_sG, o_sG):
            _gpaw.pw_precond(G2_G, r_G, ekin, o_G)


class SpinorPWHamiltonian(Hamiltonian):
    def apply(self,
              vt_xR: UniformGridFunctions,
              psit_nsG: PlaneWaveExpansions,
              out: PlaneWaveExpansions,
              spin: int):
        out_nsG = out
        pw = psit_nsG.desc

        if pw.qspiral_v is None:
            np.multiply(pw.ekin_G, psit_nsG.data, out_nsG.data)
        else:
            for s, sign in enumerate([1, -1]):
                ekin_G = 0.5 * ((pw.G_plus_k_Gv +
                                 0.5 * sign * pw.qspiral_v)**2).sum(1)
                np.multiply(ekin_G, psit_nsG.data[:, s], out_nsG.data[:, s])

        grid = vt_xR.desc.new(dtype=complex)

        v, x, y, z = vt_xR.data
        iy = y * 1j

        f_sR = grid.empty(2)
        g_R = grid.empty()

        for p_sG, o_sG in zip(psit_nsG, out_nsG):
            p_sG.ifft(out=f_sR)
            a, b = f_sR.data
            g_R.data = a * (v + z) + b * (x - iy)
            o_sG.data[0] += g_R.fft(pw=pw).data
            g_R.data = a * (x + iy) + b * (v - z)
            o_sG.data[1] += g_R.fft(pw=pw).data

        return out_nsG

    def create_preconditioner(self, blocksize):
        return spinor_precondition


"""
        N = len(psit_xG)
        S = self.gd.comm.size

        vt_R = self.gd.collect(ham.vt_sG[kpt.s], broadcast=True)
        Q_G = self.pd.Q_qG[kpt.q]
        T_G = 0.5 * self.pd.G2_qG[kpt.q]

        for n1 in range(0, N, S):
            n2 = min(n1 + S, N)
            psit_G = self.pd.alltoall1(psit_xG[n1:n2], kpt.q)
            with self.timer('HMM T'):
                np.multiply(T_G, psit_xG[n1:n2], Htpsit_xG[n1:n2])
            if psit_G is not None:
                psit_R = self.pd.ifft(psit_G, kpt.q, local=True, safe=False)
                psit_R *= vt_R
                self.pd.fftplan.execute()
                vtpsit_G = self.pd.tmp_Q.ravel()[Q_G]
            else:
                vtpsit_G = self.pd.tmp_G
            self.pd.alltoall2(vtpsit_G, kpt.q, Htpsit_xG[n1:n2])

        ham.xc.apply_orbital_dependent_hamiltonian(
            kpt, psit_xG, Htpsit_xG, ham.dH_asp)
"""



from myqueue.workflow import run


def workflow():
    with run(script='h2_gs.py'):
        run(script='h2_diss.py', cores=8)
    with run(script='graphene_h_gs.py', cores=8):
        run(script='graphene_h_prop.py', cores=32, tmax='2h')



from ase import Atoms
from gpaw import GPAW

a = 4.0
b = a / 2**0.5
L = 11.0

# Set up 2-layer 2x2 (100) Al-slab:
slab = Atoms('Al2',
             positions=[(0, 0, 0),
                        (b / 2, b / 2, -a / 2)],
             cell=(b, b, L),
             pbc=True)
slab *= (2, 2, 1)

if True:
    # Add adsorbate:
    slab += Atoms('H', [(b, b, 1.55)])

slab.center(axis=2)

calc = GPAW(mode='pw', kpts=(4, 4, 1))
slab.calc = calc
slab.get_potential_energy()
calc.write('HAl100.gpw', 'all')



import ase.io
from ase.units import Pascal, m
from ase.optimize import BFGS
from ase.neb import interpolate, DyNEB

from gpaw.solvation.sjm import SJM, SJMPower12Potential
from gpaw.solvation import (
    EffectivePotentialCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction)


def make_calculator(index):
    # Solvated jellium parameters.
    sj = {'target_potential': 4.2,
          'tol': 0.01,
          'excess_electrons': -0.01232,  # guess from previous
          'slope': -50.}  # guess from previous
    # Implicit solvent parameters (to SolvationGPAW).
    epsinf = 78.36  # dielectric constant of water at 298 K
    gamma = 18.4 * 1e-3 * Pascal * m
    cavity = EffectivePotentialCavity(
        effective_potential=SJMPower12Potential(H2O_layer=True),
        temperature=298.15,  # K
        surface_calculator=GradientSurface())
    dielectric = LinearDielectric(epsinf=epsinf)
    interactions = [SurfaceInteraction(surface_tension=gamma)]
    calc = SJM(
        # General GPAW parameters.
        txt=f'gpaw-{index:d}.txt',
        gpts=(16, 16, 136),
        kpts=(9, 9, 1),
        xc='PBE',
        maxiter=1000,
        # Solvated jellium parameters.
        sj=sj,
        # Implicit solvent parameters.
        cavity=cavity,
        dielectric=dielectric,
        interactions=interactions)
    return calc


# Create the band of images, attaching a calc to each.
initial = ase.io.read('qn-Au111-H-sim-1.traj')
final = ase.io.read('qn-Au111-H-hollow-1.traj')
images = [initial]
for index in range(5):
    images += [initial.copy()]
    images[-1].calc = make_calculator(index + 1)
images += [final]
interpolate(images)

# Create and relax the DyNEB.
neb = DyNEB(images)
opt = BFGS(neb, logfile='neb.log', trajectory='neb.traj')
opt.run(fmax=0.05)



"""DFT ground state calculation of Co(hcp)."""

# Script modules
from gpaw import GPAW, PW, FermiDirac
from ase.build import bulk

# ---------- Inputs ---------- #

# Crystal structure
a = 2.5071
c = 4.0695
mm = 1.67

# Calculator parameters
xc = 'LDA'
pw = 800  # eV
# For a hcp crystal, a k-point grid that samples a multiple of 6 k-points
# along all axes will support calculations at the high-symmetry points M, K and
# A. For this tutorial, we choose N_k^(1/3)=24.
kpts = 24  # final grid is (kpts, kpts, kpts)
nbands_mft = 2 * 6  # natoms * (4s + 3d)
nbands_gs = nbands_mft + 2 * 4  # + natoms * (4p + 5s)
conv = {'density': 1.e-8,
        'forces': 1.e-8,
        'bands': nbands_mft}  # Converge only the bands needed subsequently
occw = 0.001  # Fermi temperature in eV

# ---------- Script ---------- #

# Set up crystal structure
atoms = bulk('Co', 'hcp', a=a, c=c)
atoms.set_initial_magnetic_moments([mm, mm])
atoms.center()

calc = GPAW(xc='LDA',
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            nbands=nbands_gs,
            convergence=conv,
            occupations=FermiDirac(occw),
            txt='Co_gs.txt')

atoms.calc = calc
atoms.get_potential_energy()

# Save the ground state w.o. the Kohn-Sham orbitals
calc.write('Co.gpw')



from gpaw.tddft import TDDFT, photoabsorption_spectrum
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField

# Calculate photoabsorption spectrum as usual
folding = 'Gauss'
width = 0.1
e_min = 0.0
e_max = 4.0
photoabsorption_spectrum('na2_td_dm.dat', 'na2_td_spectrum_x.dat',
                         folding=folding, width=width,
                         e_min=e_min, e_max=e_max, delta_e=1e-2)

# Load TDDFT object
td_calc = TDDFT('na2_td.gpw')

# Load InducedField object
ind = TDDFTInducedField(filename='na2_td.ind',
                        paw=td_calc)

# Calculate induced electric field
ind.calculate_induced_field(gridrefinement=2, from_density='comp')

# Save induced electric field
ind.write('na2_td_field.ind', mode='all')



from math import pi

import numpy as np
import pytest
from scipy.special import expn
import ase.units as units

from gpaw.grid_descriptor import GridDescriptor
from gpaw.hyperfine import (hyperfine_parameters, paw_correction, smooth_part,
                            integrate, alpha, g_factor_e, core_contribution)
from gpaw import GPAW
from gpaw.atom.aeatom import AllElectronAtom
from gpaw.atom.radialgd import RadialGridDescriptor
from gpaw.lfc import LFC
from gpaw.setup import create_setup
from gpaw.xc import XC


@pytest.mark.serial
def test_thomson_integral():
    dr = 0.0001
    rgd = RadialGridDescriptor.new('equidistant', dr, 16000)
    beta = 0.0
    rT = 0.001
    n0_g = np.exp(-rgd.r_g)

    n0 = integrate(n0_g, rgd, rT, beta)
    ref = expn(2, rT / 2) * np.exp(rT / 2)
    assert n0 == pytest.approx(ref, 1e-5)


@pytest.mark.serial
def test_thomson_integral2():
    dr = 0.0001
    rgd = RadialGridDescriptor.new('equidistant', dr, 66000)
    rT = 0.001
    beta = 0.001
    n0_g = rgd.zeros()
    n0_g[1:] = rgd.r_g[1:]**-beta
    n0_g[0] = np.nan

    n0 = integrate(n0_g, rgd, rT, beta)
    ref = (2 / rT)**beta * np.pi * rT / np.sin(np.pi * rT)
    assert n0 == pytest.approx(ref, 1e-4)


class Setup:
    def __init__(self):
        self.rgd = RadialGridDescriptor.new('equidistant', 0.01, 400)
        self.data = Data(self.rgd)
        self.l_j = [0, 1]
        self.Z = 1
        self.Nc = 0


class Data:
    def __init__(self, rgd: RadialGridDescriptor):
        self.phit_jg = [np.exp(-(rgd.r_g / 0.5)**2),
                        np.exp(-(rgd.r_g / 0.5)**2) * rgd.r_g]
        self.phi_jg = [rgd.zeros(), rgd.zeros()]


@pytest.fixture
def things():
    setup = Setup()

    n = 40
    L = n * 0.1
    gd = GridDescriptor((n, n, n), (L, L, L), (1, 1, 1))

    splines = []
    for j, phit_g in enumerate(setup.data.phit_jg):
        splines.append(setup.rgd.spline(phit_g, 1.5, l=j, points=101))

    lfc = LFC(gd, [splines])

    spos_ac = np.zeros((1, 3)) + 0.5
    lfc.set_positions(spos_ac)

    return gd, lfc, setup, spos_ac


@pytest.mark.serial
def test_gaussian(things):
    gd, lfc, setup, spos_ac = things
    spin_density_R = gd.zeros()
    lfc.add(spin_density_R, {0: np.array([1.0, 0, 0, 0])})
    spin_density_R *= spin_density_R

    W_avv = smooth_part(spin_density_R, gd, spos_ac)
    print(W_avv)
    assert abs(W_avv[0] - np.eye(3) * W_avv[0, 0, 0]).max() < 1e-7

    density_sii = np.zeros((2, 4, 4))
    density_sii[0, 0, 0] = 1.0
    W1_vv = paw_correction(density_sii, setup)
    print(W1_vv)
    assert abs(W_avv[0] + W1_vv).max() < 1e-7


@pytest.mark.serial
def test_gaussian2(things):
    gd, lfc, setup, spos_ac = things
    spin_density_R = gd.zeros()
    lfc.add(spin_density_R, {0: np.array([0, 0, 1.0, 0])})
    spin_density2_R = gd.zeros()
    lfc.add(spin_density2_R, {0: np.array([0, 1.0, 0, 0])})
    spin_density_R = spin_density_R * spin_density2_R

    W_avv = smooth_part(spin_density_R, gd, spos_ac)
    print(W_avv)
    assert abs(W_avv[0] - np.array([[0, 0, 0],
                                    [0, 0, 1],
                                    [0, 1, 0]]) * W_avv[0, 1, 2]).max() < 1e-7

    density_sii = np.zeros((2, 4, 4))
    density_sii[0, 2, 1] = 1.0
    W1_vv = paw_correction(density_sii, setup)
    print(W1_vv)
    assert abs(W_avv[0] + W1_vv).max() < 1e-6


g_factor_proton = 5.586


@pytest.mark.serial
def test_h(gpw_files):
    calc = GPAW(gpw_files['h_pw'])
    A_vv = hyperfine_parameters(calc)[0] * g_factor_proton
    print(A_vv)

    energy = (2 / 3 * alpha**2 * units.Ha * units._me / units._mp *
              g_factor_e * g_factor_proton)  # in eV
    frequency = energy * units._e / units._hplanck  # Hz
    wavelength = units._c / frequency  # meters
    assert wavelength == pytest.approx(0.211, abs=0.001)

    energy *= 0.94  # density at nucleus is slightly below 1/pi for PBE
    print(energy)
    assert abs(A_vv - np.eye(3) * energy).max() < 1e-7


def thomson():
    """Analytic integrals for testing."""
    from sympy import var, integrate, oo, E, expint
    x, a, b = var('x, a, b')
    print(integrate(E**(-b * x) / (1 + x)**2, (x, 0, oo)))
    print(expint(2, 1.0))


def test_hyper_core():
    """Test polarization of "frozen" nitrogen-1s core."""
    # Calculate 1s spin-density from 2s and 2p polarizarion:
    setup = create_setup('N')
    xc = XC('LDA')
    D_sii = np.zeros((2, 13, 13))
    D_sii[:, 0, 0] = 1.0
    D_sii[0, 1:4, 1:4] = np.eye(3)
    spin_density1_g = core_contribution(D_sii, setup, xc)

    assert setup.rgd.integrate(spin_density1_g) == pytest.approx(0, abs=1e-11)

    # Do all-electron calculation:
    aea = AllElectronAtom('N', spinpol=True)
    aea.initialize()
    aea.run()
    aea.scalar_relativistic = True
    aea.refine()

    # 1s:
    spin_density2_g = (aea.channels[0].phi_ng[0]**2 -
                       aea.channels[1].phi_ng[0]**2) / (4 * pi)
    # 2s:
    spin_density3_g = (aea.channels[0].phi_ng[1]**2 -
                       aea.channels[1].phi_ng[1]**2) / (4 * pi)

    assert aea.rgd.integrate(spin_density2_g) == pytest.approx(0, abs=1e-11)
    assert aea.rgd.integrate(spin_density3_g) == pytest.approx(0, abs=1e-11)

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(setup.rgd.r_g, spin_density1_g)
        plt.plot(aea.rgd.r_g, spin_density2_g)
        plt.show()

    assert spin_density1_g[0] == pytest.approx(spin_density2_g[0], abs=0.04)



from ase import Atoms
from gpaw import GPAW

L = 6
name = 'N2'
a = Atoms('N2',
          [(L / 2 + 1.098 / 2, L / 2, L / 2),
           (L / 2 - 1.098 / 2, L / 2, L / 2)],
          cell=(L, L, L), pbc=False)

calc = GPAW(h=0.22,
            xc='PBE',
            convergence={'eigenstates': 1.0e-7},
            txt=name + '.txt',
            eigensolver='rmm-diis')
a.calc = calc
e_n2 = a.get_potential_energy()
n2t = calc.get_xc_difference('TPSS')
n2rt = calc.get_xc_difference('revTPSS')

a.calc.set(xc='TPSS')
e_n2t = a.get_potential_energy()

a.calc.set(xc='revTPSS')
e_n2rt = a.get_potential_energy()

name = 'N'
b = Atoms('N', [(L / 2, L / 2, L / 2)], magmoms=[3],
          cell=(L, L, L), pbc=False)

calc = GPAW(h=0.22,
            xc='PBE',
            convergence={'eigenstates': 1.0e-7},
            txt=name + '.txt',
            eigensolver='rmm-diis',
            hund=True)
b.calc = calc
e_n = b.get_potential_energy()
nt = calc.get_xc_difference('TPSS')
nrt = calc.get_xc_difference('revTPSS')

b.calc.set(xc='TPSS')
e_nt = b.get_potential_energy()

b.calc.set(xc='revTPSS')
e_nrt = b.get_potential_energy()

print('Atm. Experiment  ', -228.5)
print('Atm. PBE         ', (e_n2 - 2 * e_n) * 23.06)
print('Atm. TPSS(nsc)   ', ((e_n2 + n2t) - 2 * (e_n + nt)) * 23.06)
print('Atm. TPSS        ', (e_n2t - 2 * e_nt) * 23.06)
print('Atm. revTPSS(nsc)', ((e_n2 + n2rt) - 2 * (e_n + nrt)) * 23.06)
print('Atm. revTPSS     ', (e_n2rt - 2 * e_nrt) * 23.06)



from ase.io import read
from gpaw import GPAW, FermiDirac, Mixer, PoissonSolver
from gpaw.poisson_moment import MomentCorrectionPoissonSolver

# Read the structure from the xyz file
atoms = read('Ag55.xyz')
atoms.center(vacuum=6.0)

# Increase the accuracy of density for ground state
convergence = {'density': 1e-12}

# Use occupation smearing and weak mixer to facilitate convergence
occupations = FermiDirac(25e-3)
mixer = Mixer(0.02, 5, 1.0)

# Parallelzation settings
parallel = {'sl_auto': True, 'domain': 2, 'augment_grids': True}

# Apply multipole corrections for monopole and dipoles
poissonsolver = MomentCorrectionPoissonSolver(poissonsolver=PoissonSolver(),
                                              moment_corrections=1 + 3)

# Ground-state calculation
calc = GPAW(mode='lcao', xc='GLLBSC', h=0.3, nbands=360,
            setups={'Ag': '11'},
            basis={'Ag': 'pvalence.dz', 'default': 'dzp'},
            convergence=convergence, poissonsolver=poissonsolver,
            occupations=occupations, mixer=mixer, parallel=parallel,
            maxiter=1000,
            txt='gs.out')
atoms.calc = calc
atoms.get_potential_energy()
calc.write('gs.gpw', mode='all')



# General modules
import numpy as np

# GPAW modules
import gpaw.mpi as mpi
from gpaw.response.chiks import ChiKS
from gpaw.response.kxc import PlaneWaveAdiabaticFXC
from gpaw.response.site_kernels import SiteKernels
from gpaw.response.susceptibility import symmetrize_reciprocity
from gpaw.xc import XC

# ASE modules
from ase.units import Hartree


class IsotropicExchangeCalculator:
    r"""Calculator class for the Heisenberg exchange constants

    _           2
    J^ab(q) = - ‾‾ B^(xc†) K^(a†)(q) χ_KS^('+-)(q) K^b(q) B^(xc)
                V0

    calculated for an isotropic system in a plane wave representation using
    the magnetic force theorem within second order perturbation theory, see
    [arXiv:2204.04169].

    Entering the formula for the isotropic exchange constant at wave vector q
    between sublattice a and b is the unit cell volume V0, the functional
    derivative of the (LDA) exchange-correlation energy with respect to the
    magnitude of the magnetization B^(xc), the sublattice site kernels K^a(q)
    and K^b(q) as well as the reactive part of the static transverse magnetic
    susceptibility of the Kohn-Sham system χ_KS^('+-)(q).

    The site kernels encode the partitioning of real space into sites of the
    Heisenberg model. This is not a uniquely defined procedure, why the user
    has to define them externally through the SiteKernels interface."""

    def __init__(self, chiks):
        """Construct the IsotropicExchangeCalculator object

        Parameters
        ----------
        chiks : ChiKS
            ChiKS calculator object
        """
        assert isinstance(chiks, ChiKS)
        # Check that chiks has the assumed properties
        assumed_props = dict(
            gammacentered=True,
            kpointintegration='point integration',
            nblocks=1
        )
        for key, item in assumed_props.items():
            assert getattr(chiks, key) == item,\
                f'Expected chiks.{key} == {item}. Got: {getattr(chiks, key)}'

        self.chiks = chiks

        # Initialize the B^(xc) calculator
        # Once the response context object is ready, the user should be allowed
        # to supply the Bxc_calc themselves. This will expose the rshe
        # arguments to the user, which is not the case at present. XXX
        self.Bxc_calc = PlaneWaveBxc(self.chiks.calc,
                                     world=self.chiks.world,
                                     txt=self.chiks.fd,
                                     timer=self.chiks.timer)

        # Bxc field buffer
        self._Bxc_G = None

        # chiksr buffer
        self.currentq_c = None
        self._pd = None
        self._chiksr_GG = None

    def __call__(self, q_c, site_kernels, txt=None):
        """Calculate the isotropic exchange constants for a given wavevector.

        Parameters
        ----------
        q_c : nd.array
            Wave vector q in relative coordinates
        site_kernels : SiteKernels
            Site kernels instance defining the magnetic sites of the crystal
        txt : str
            Separate file to store the chiks calculation output in (optional).
            If not supplied, the output will be written to the standard text
            output location specified when initializing chiks.

        Returns
        -------
        J_abp : nd.array (dtype=complex)
            Isotropic Heisenberg exchange constants between magnetic sites a
            and b for all the site partitions p given by the site_kernels.
        """
        assert isinstance(site_kernels, SiteKernels)

        # Get ingredients
        Bxc_G = self.get_Bxc()
        pd, chiksr_GG = self.get_chiksr(q_c, txt=txt)
        V0 = pd.gd.volume

        # Allocate an array for the exchange constants
        nsites = site_kernels.nsites
        J_pab = np.empty(site_kernels.shape + (nsites,), dtype=complex)

        # Compute exchange coupling
        for J_ab, K_aGG in zip(J_pab, site_kernels.calculate(pd)):
            for a in range(nsites):
                for b in range(nsites):
                    J = np.conj(Bxc_G) @ np.conj(K_aGG[a]).T @ chiksr_GG \
                        @ K_aGG[b] @ Bxc_G
                    J_ab[a, b] = - 2. * J / V0

        # Transpose to have the partitions index last
        J_abp = np.transpose(J_pab, (1, 2, 0))

        return J_abp * Hartree  # Convert from Hartree to eV

    def get_Bxc(self):
        """Get B^(xc)_G from buffer."""
        if self._Bxc_G is None:  # Calculate, if buffer is empty
            self._Bxc_G = self._calculate_Bxc()

        return self._Bxc_G

    def _calculate_Bxc(self):
        """Use the PlaneWaveBxc calculator to calculate the plane wave
        coefficients B^xc_G"""
        # Create a plane wave descriptor encoding the plane wave basis. Input
        # q_c is arbitrary, since we are assuming that chiks.gammacentered == 1
        pd0 = self.chiks.get_PWDescriptor([0., 0., 0.])

        return self.Bxc_calc(pd0)

    def get_chiksr(self, q_c, txt=None):
        """Get χ_KS^('+-)(q) from buffer."""
        q_c = np.asarray(q_c)
        if self.currentq_c is None or not np.allclose(q_c, self.currentq_c):
            # Calculate chiks for any new q-point or if buffer is empty
            self.currentq_c = q_c
            self._pd, self._chiksr_GG = self._calculate_chiksr(q_c, txt=txt)

        return self._pd, self._chiksr_GG

    def _calculate_chiksr(self, q_c, txt=None):
        r"""Use the ChiKS calculator to calculate the reactive part of the
        static Kohn-Sham susceptibility χ_KS^('+-)(q).

        First, the dynamic Kohn-Sham susceptibility

                                 __  __
                              1  \   \        f_nk↑ - f_mk+q↓
        χ_KS,GG'^+-(q,ω+iη) = ‾  /   /  ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                              V  ‾‾  ‾‾ ħω - (ε_mk+q↓ - ε_nk↑) + iħη
                                 k  n,m
                                        x n_nk↑,mk+q↓(G+q) n_mk+q↓,nk↑(-G'-q)

        is calculated in the static limit ω=0. Then, the reactive part (see
        [PRB 103, 245110 (2021)]) is extracted,

                              1
        χ_KS,GG'^(+-')(q,ω) = ‾ [χ_KS,GG'^+-(q,ω+iη) + χ_KS,-G'-G^-+(-q,-ω+iη)]
                              2

                              1
                            = ‾ [χ_KS,GG'^+-(q,ω+iη) + χ_KS,G'G^(+-*)(q,ω+iη)]
                              2

        where it was used that n^+(r) and n^-(r) are each others Hermitian
        conjugates to reach the last equality.
        """
        frequencies = [0.]
        pd, chiks_wGG = self.chiks.calculate(q_c, frequencies,
                                             spincomponent='+-',
                                             txt=txt)
        symmetrize_reciprocity(pd, chiks_wGG)

        # Take the reactive part
        chiksr_GG = 1 / 2. * (chiks_wGG[0] + np.conj(chiks_wGG[0]).T)

        return pd, chiksr_GG


class PlaneWaveBxc(PlaneWaveAdiabaticFXC):
    """Calculator class for the plane wave coefficients of B^(xc)

               /
    B^(xc)_G = |dr B^(xc)(r) e^(-iG.r)
               /
                V0

    where V0 is the cell volume and

                δE_xc[n,m]   1
    B^(xc)(r) = ‾‾‾‾‾‾‾‾‾‾ = ‾ [V_xc^↑(r) - V_xc^↓(r)]
                  δm(r)      2

    in the local spin-density approximation for a collinear system."""

    def __init__(self, gs,
                 world=mpi.world, txt='-', timer=None,
                 rshelmax=-1, rshewmin=1.e-8):  # Overwrites rshewmin default
        """Construct the calculator based on functionality to compute fxc
        kernels. This is a temporary hack to leverage the PAW functionality
        of that code, but implies a significant computational overhead.

        Parameters
        ----------
        gs, world, txt, timer : see FXC
        rshelmax, rshewmin : see PlaneWaveAdiabaticFXC
        """
        PlaneWaveAdiabaticFXC.__init__(self, gs, '',
                                       world=world, txt=txt, timer=timer,
                                       rshelmax=rshelmax, rshewmin=rshewmin)

    def __call__(self, pd):
        """Calculate the plane wave components of Bxc"""
        # Use the fxc kernel functionality to compute B^(xc)_(G-G') / V0,
        # see [PRB 103, 245110 (2021)]
        Bxc_GG = self.calculate(pd)

        # Extract B^(xc)_G as the first column of the "kernel", renormalizing
        # by the cell volume
        V0 = pd.gd.volume
        Bxc_G = V0 * Bxc_GG[:, 0]

        return Bxc_G

    def _add_fxc(self, gd, n_sG, fxc_G):
        """This function defines the "fxc kernel" to Fourier transform."""
        self._add_Bxc(gd, n_sG, fxc_G)

    def _add_Bxc(self, gd, n_sG, Bxc_G):
        """Calculate Bxc in real space and add it to the array Bxc_G (here
        G denotes the real space grid points)."""
        # Allocate an array for the spin-dependent xc potential on the real
        # space grid
        v_sG = np.zeros(np.shape(n_sG))

        # Calculate the spin-dependent potential
        xc = XC('LDA')
        xc.calculate(gd, n_sG, v_sg=v_sG)

        # Add B^(xc) in real space to the output array
        Bxc_G += (v_sG[0] - v_sG[1]) / 2



from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters

# Generate setups with 0.5, 1.0, 0.0 core holes in 1s
elements = ['O', 'C', 'N']
coreholes = [0.5, 1.0, 0.0]
names = ['hch1s', 'fch1s', 'xes1s']
functionals = ['LDA', 'PBE']

for el in elements:
    for name, ch in zip(names, coreholes):
        for funct in functionals:
            g = Generator(el, scalarrel=True, xcname=funct,
                          corehole=(1, 0, ch), nofiles=True)
            g.run(name=name, **parameters[el])



# web-page: simultaneous.png

from matplotlib import pyplot, gridspec


def makefig():
    fig = pyplot.figure(figsize=(6.5, 6.5))
    grid = gridspec.GridSpec(2, 1, fig, right=0.95, top=0.95)
    smallgrid = gridspec.GridSpecFromSubplotSpec(2, 1, grid[0], hspace=0.)
    ax0 = fig.add_subplot(smallgrid[0])
    ax1 = fig.add_subplot(smallgrid[1], sharex=ax0)
    smallgrid = gridspec.GridSpecFromSubplotSpec(2, 1, grid[1], hspace=0.)
    ax2 = fig.add_subplot(smallgrid[0], sharex=ax0)
    ax3 = fig.add_subplot(smallgrid[1], sharex=ax0)
    return fig, ax0, ax1, ax2, ax3


def get_data(mode):
    with open(f'Au111-H-{mode}.txt', 'r') as f:
        lines = f.read().splitlines()
    potentials = []
    equilibrateds = []
    fmaxs = []
    for line in lines:
        if 'Potential found to be' in line:
            potentials += [float(line.split('to be')[-1].split('V (with')[0])]
        if 'Potential is within tolerance. Equilibrated.' in line:
            diff = [False] * (len(potentials) - len(equilibrateds))
            diff[-1] = True
            equilibrateds.extend(diff)
    with open(f'qn-Au111-H-{mode}.log', 'r') as f:
        lines = f.read().splitlines()
    for line in lines:
        if 'BFGS:' in line:
            fmaxs += [float(line.split()[-1])]
    return potentials, equilibrateds, fmaxs


def plot_data(potentials, equilibrateds, fmaxs, axes):
    fmax_indices = []
    axes[0].plot(potentials, 'o-', color='k', markerfacecolor='w')
    for index, (potential, equilibrated) in enumerate(zip(potentials,
                                                          equilibrateds)):
        if equilibrated:
            axes[0].plot(index, potential, 'og', markeredgecolor='k')
            fmax_indices += [index]
    axes[1].plot(fmax_indices, fmaxs, 'ko-')


def prune_duplicate(potentials, equilibrateds):
    """Due to logging, there's a duplicate potential reading when
    we switched from simultaneous to sequential. Remove it, and also
    return the image number of the switch."""
    for index, potential in enumerate(potentials):
        if index > 0 and potential == potentials[index - 1]:
            duplicate = index
    del potentials[duplicate]
    del equilibrateds[duplicate]
    return duplicate


def shade_plot(ax, xstart):
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    ax.fill_between(x=[xstart, xlim[-1]], y1=[ylim[0]] * 2, y2=[ylim[1]] * 2,
                    color='0.8')
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)


fig, ax0, ax1, ax2, ax3 = makefig()
potentials, equilibrateds, fmaxs = get_data('seq')
plot_data(potentials, equilibrateds, fmaxs, [ax0, ax1])
potentials, equilibrateds, fmaxs = get_data('sim')
duplicate = prune_duplicate(potentials, equilibrateds)
plot_data(potentials, equilibrateds, fmaxs, [ax2, ax3])

ax3.set_xlabel('DFT calculation')
ax0.text(0.20, 0.75, 'sequential', transform=ax0.transAxes)
ax2.text(0.20, 0.75, 'simultaneous', transform=ax2.transAxes)
ax2.text(0.75, 0.75, '(sequential)', transform=ax2.transAxes)
shade_plot(ax2, duplicate - 0.5)
shade_plot(ax3, duplicate - 0.5)
for ax in [ax0, ax2]:
    ax.set_ylabel(r'$\phi$, V')
for ax in [ax1, ax3]:
    ax.set_ylabel(r'max force, eV/$\mathrm{\AA}$')

fig.savefig('simultaneous.png')



import pytest

from gpaw import GPAW, LCAO
from ase import Atoms
import numpy as np
from gpaw.directmin.etdm import ETDM


def test_directmin_lcao(in_tmp_dir):
    """
    test exponential transformation
    direct minimization method for KS-DFT in LCAO
    :param in_tmp_dir:
    :return:
    """

    # Water molecule:
    d = 0.9575
    t = np.pi / 180 * 104.51
    H2O = Atoms('OH2',
                positions=[(0, 0, 0),
                           (d, 0, 0),
                           (d * np.cos(t), d * np.sin(t), 0)])
    H2O.center(vacuum=5.0)

    calc = GPAW(mode=LCAO(),
                basis='dzp',
                occupations={'name': 'fixed-uniform'},
                eigensolver='etdm',
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                symmetry='off'
                )
    H2O.calc = calc
    e = H2O.get_potential_energy()
    f = H2O.get_forces()

    assert e == pytest.approx(-13.643156256566218, abs=1.0e-4)
    f2 = np.array([[-1.11463, -1.23723, 0.0],
                   [1.35791, 0.00827, 0.0],
                   [-0.34423, 1.33207, 0.0]])
    assert f2 == pytest.approx(f, abs=1e-2)

    calc.write('h2o.gpw', mode='all')
    from gpaw import restart
    H2O, calc = restart('h2o.gpw', txt='-')
    H2O.positions += 1.0e-6
    f3 = H2O.get_forces()
    niter = calc.get_number_of_iterations()

    assert niter == pytest.approx(3, abs=1)
    assert f2 == pytest.approx(f3, abs=1e-2)

    calc.set(eigensolver=ETDM(
        representation='u-invar', matrix_exp='egdecomp-u-invar',
        need_init_orbs=False,
        linesearch_algo={'name': 'max-step'}
    ))
    e = H2O.get_potential_energy()
    niter = calc.get_number_of_iterations()
    assert e == pytest.approx(-13.643156256566218, abs=1.0e-4)
    assert niter == pytest.approx(3, abs=1)



"""Calculate the vibrational modes of a H2O molecule."""

from ase.vibrations import Vibrations
from gpaw import GPAW

h2o = GPAW('h2o.gpw', txt=None).get_atoms()

# Create vibration calculator
vib = Vibrations(h2o)
vib.run()
vib.summary(method='frederiksen')

# Make trajectory files to visualize normal modes:
for mode in range(9):
    vib.write_mode(mode)



def workflow():
    from myqueue.workflow import run
    runs = [run(script='ferro.py', cores=4, tmax='15m'),
            run(script='anti.py', cores=4, tmax='15m'),
            run(script='non.py', cores=2, tmax='15m')]
    run(script='PBE.py', deps=runs)
    for r, name in zip(runs, ['ferro', 'anti', 'non']):
        run(script='../../electronic/dos/dos.py',
            args=[f'{name}.gpw'], deps=[r])
    run(script='../../electronic/dos/pdos.py', deps=[runs[0]])



import numpy as np
from time import ctime

from gpaw.utilities import convert_string_to_fd
from ase.utils.timing import timer

from gpaw.utilities.blas import mmmx
from gpaw.response.kslrf import PlaneWaveKSLRF
from gpaw.response.kspair import PlaneWavePairDensity


class ChiKS(PlaneWaveKSLRF):
    r"""Class calculating the four-component Kohn-Sham susceptibility tensor,
    see [PRB 103, 245110 (2021)]. For collinear systems, the susceptibility
    tensor is defined as:
                        __  __   __
                     1  \   \    \
    chiKSmunu(q,w) = ‾  /   /    /   (f_nks - f_n'k+qs') smu_ss' snu_s's
                     V  ‾‾  ‾‾   ‾‾
                        k  n,s  n',s'
                                       n_nks,n'k+qs'(G+q) n_n'k+qs',nks(-G'-q)
                                     x ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                                         hw - (eps_n'k+qs'-eps_nks) + ih eta

    where the matrix elements

    n_nks,n'k+qs'(G+q) = <nks| e^-i(G+q)r |n'k+qs'>_V0

    are unit cell normalized plane wave pair densities of each transition.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the chiKS object in plane wave mode."""
        PlaneWaveKSLRF.__init__(self, *args, **kwargs)

        # Susceptibilities use pair densities as matrix elements
        self.pme = PlaneWavePairDensity(self.kspair)

        # The class is calculating one spin component at a time
        self.spincomponent = None

    @timer('Calculate the Kohn-Sham susceptibility')
    def calculate(self, q_c, frequencies, spincomponent='all',
                  A_x=None, txt=None):
        """Calculate a component of the susceptibility tensor.

        Parameters
        ----------
        spincomponent : str or int
            What susceptibility should be calculated?
            Currently, '00', 'uu', 'dd', '+-' and '-+' are implemented.
            'all' is an alias for '00', kept for backwards compability
            Likewise 0 or 1, can be used for 'uu' or 'dd'
        """
        # Initiate new call-output file, if supplied
        if txt is not None:
            self.cfd = convert_string_to_fd(txt, self.world)
        # Print to output file(s)
        if str(self.fd) != str(self.cfd) or txt is not None:
            print('Calculating Kohn-Sham susceptibility with '
                  f'q_c={q_c} and spincomponent={spincomponent}',
                  file=self.fd)
            print(ctime(), file=self.fd)

        # Analyze the requested spin component
        self.spincomponent = spincomponent
        spinrot = get_spin_rotation(spincomponent)

        return PlaneWaveKSLRF.calculate(self, q_c, frequencies,
                                        spinrot=spinrot, A_x=A_x)

    @timer('Add integrand to chiks_wGG')
    def add_integrand(self, kskptpair, weight, A_x):
        r"""Use PairDensity object to calculate the integrand for all relevant
        transitions of the given k-point pair (k,k+q).

        Depending on the bandsummation parameter, the integrand of the
        collinear four-component Kohn-Sham susceptibility tensor is calculated
        as:

        bandsummation: double

                   __
                   \  smu_ss' snu_s's (f_nks - f_n'k's')
        (...)_k =  /  ---------------------------------- n_kt(G+q) n_kt^*(G'+q)
                   ‾‾ hw - (eps_n'k's'-eps_nks) + ih eta
                   t

        where n_kt(G+q) = n_nks,n'k+qs'(G+q) and

        bandsummation: pairwise (using spin-conserving time-reversal symmetry)

                    __ /
                    \  | smu_ss' snu_s's (f_nks - f_n'k's')
        (...)_k =   /  | ----------------------------------
                    ‾‾ | hw - (eps_n'k's'-eps_nks) + ih eta
                    t  \
                                                       \
                    smu_s's snu_ss' (f_nks - f_n'k's') |
        -delta_n'>n ---------------------------------- | n_kt(G+q) n_kt^*(G'+q)
                    hw + (eps_n'k's'-eps_nks) + ih eta |
                                                       /

        The integrand is added to the output array A_x multiplied with the
        supplied k-point weight.
        """
        # Get data, distributed in memory
        # Get bands and spins of the transitions
        n1_t, n2_t, s1_t, s2_t = kskptpair.get_transitions()
        # Get (f_n'k's' - f_nks), (eps_n'k's' - eps_nks) and the pair densities
        df_t, deps_t, n_tG = kskptpair.df_t, kskptpair.deps_t, kskptpair.n_tG

        x_wt = weight * self.get_temporal_part(n1_t, n2_t,
                                               s1_t, s2_t, df_t, deps_t)

        myslice = self.blocks1d.myslice

        if self.bundle_integrals:
            # Specify notation
            A_GwmyG = A_x

            x_tw = np.ascontiguousarray(x_wt.T)
            n_Gt = np.ascontiguousarray(n_tG.T)

            with self.timer('Set up ncc and nx'):
                ncc_Gt = n_Gt.conj()
                n_tmyG = n_tG[:, myslice]
                nx_twmyG = x_tw[:, :, np.newaxis] * n_tmyG[:, np.newaxis, :]

            with self.timer('Perform sum over t-transitions of ncc * nx'):
                mmmx(1.0, ncc_Gt, 'N', nx_twmyG, 'N',
                     1.0, A_GwmyG)  # slow step
        else:
            # Specify notation
            A_wmyGG = A_x

            with self.timer('Set up ncc and nx'):
                ncc_tG = n_tG.conj()
                n_myGt = np.ascontiguousarray(n_tG[:, myslice].T)
                nx_wmyGt = x_wt[:, np.newaxis, :] * n_myGt[np.newaxis, :, :]

            with self.timer('Perform sum over t-transitions of ncc * nx'):
                for nx_myGt, A_myGG in zip(nx_wmyGt, A_wmyGG):
                    mmmx(1.0, nx_myGt, 'N', ncc_tG, 'N',
                         1.0, A_myGG)  # slow step

    @timer('Get temporal part')
    def get_temporal_part(self, n1_t, n2_t, s1_t, s2_t, df_t, deps_t):
        """Get the temporal part of the susceptibility integrand."""
        _get_temporal_part = self.create_get_temporal_part()
        return _get_temporal_part(n1_t, n2_t, s1_t, s2_t, df_t, deps_t)

    def create_get_temporal_part(self):
        """Creator component, deciding how to calculate the temporal part"""
        if self.bandsummation == 'double':
            return self.get_double_temporal_part
        elif self.bandsummation == 'pairwise':
            return self.get_pairwise_temporal_part
        raise ValueError(self.bandsummation)

    def get_double_temporal_part(self, n1_t, n2_t, s1_t, s2_t, df_t, deps_t):
        """Get:

               smu_ss' snu_s's (f_nks - f_n'k's')
        x_wt = ----------------------------------
               hw - (eps_n'k's'-eps_nks) + ih eta
        """
        # Get the right spin components
        scomps_t = get_smat_components(self.spincomponent, s1_t, s2_t)
        # Calculate nominator
        nom_t = - scomps_t * df_t  # df = f2 - f1
        # Calculate denominator
        denom_wt = self.wd.omega_w[:, np.newaxis] + 1j * self.eta\
            - deps_t[np.newaxis, :]  # de = e2 - e1

        return nom_t[np.newaxis, :] / denom_wt

    def get_pairwise_temporal_part(self, n1_t, n2_t, s1_t, s2_t, df_t, deps_t):
        """Get:
               /
               | smu_ss' snu_s's (f_nks - f_n'k's')
        x_wt = | ----------------------------------
               | hw - (eps_n'k's'-eps_nks) + ih eta
               \
                                                           \
                        smu_s's snu_ss' (f_nks - f_n'k's') |
            -delta_n'>n ---------------------------------- |
                        hw + (eps_n'k's'-eps_nks) + ih eta |
                                                           /
        """
        # Dirac delta
        delta_t = np.ones(len(n1_t))
        delta_t[n2_t <= n1_t] = 0
        # Get the right spin components
        scomps1_t = get_smat_components(self.spincomponent, s1_t, s2_t)
        scomps2_t = get_smat_components(self.spincomponent, s2_t, s1_t)
        # Calculate nominators
        nom1_t = - scomps1_t * df_t  # df = f2 - f1
        nom2_t = - delta_t * scomps2_t * df_t
        # Calculate denominators
        denom1_wt = self.wd.omega_w[:, np.newaxis] + 1j * self.eta\
            - deps_t[np.newaxis, :]  # de = e2 - e1
        denom2_wt = self.wd.omega_w[:, np.newaxis] + 1j * self.eta\
            + deps_t[np.newaxis, :]

        return nom1_t[np.newaxis, :] / denom1_wt\
            - nom2_t[np.newaxis, :] / denom2_wt


def get_spin_rotation(spincomponent):
    """Get the spin rotation corresponding to the given spin component."""
    if spincomponent is None or spincomponent == '00':
        return '0'
    elif spincomponent in ['uu', 'dd', '+-', '-+']:
        return spincomponent[-1]
    else:
        raise ValueError(spincomponent)


def get_smat_components(spincomponent, s1_t, s2_t):
    """For s1=s and s2=s', get:
    smu_ss' snu_s's
    """
    if spincomponent is None:
        spincomponent = '00'

    smatmu = smat(spincomponent[0])
    smatnu = smat(spincomponent[1])

    return smatmu[s1_t, s2_t] * smatnu[s2_t, s1_t]


def smat(spinrot):
    if spinrot == '0':
        return np.array([[1, 0], [0, 1]])
    elif spinrot == 'u':
        return np.array([[1, 0], [0, 0]])
    elif spinrot == 'd':
        return np.array([[0, 0], [0, 1]])
    elif spinrot == '-':
        return np.array([[0, 0], [1, 0]])
    elif spinrot == '+':
        return np.array([[0, 1], [0, 0]])
    else:
        raise ValueError(spinrot)



from ase.build import bulk
from gpaw import GPAW, PW


def test_pseudopotential_ah(in_tmp_dir):
    si = bulk('Si', 'diamond', a=5.5, cubic=not True)
    si.calc = GPAW(mode=PW(200),
                   setups='ah',
                   kpts=(2, 2, 2))
    si.get_forces()
    si.get_stress()
    si.calc.write('Si.gpw', 'all')
    GPAW('Si.gpw')



from ase.parallel import paropen
from gpaw.xc.rpa import RPACorrelation
import numpy as np

dw = 0.5
frequencies = np.array([dw * i for i in range(200)])
weights = len(frequencies) * [dw]
weights[0] /= 2
weights[-1] /= 2
weights = np.array(weights)

rpa = RPACorrelation('N2.gpw',
                     txt='frequency_equidistant.txt',
                     frequencies=frequencies,
                     weights=weights)

Es = rpa.calculate(ecut=[50])
Es_w = rpa.E_w

with paropen('frequency_equidistant.dat', 'w') as fd:
    for w, E in zip(frequencies, Es_w):
        print(w, E.real, file=fd)



from types import SimpleNamespace

import pytest
import numpy as np
from ase.units import Bohr as bohr

from gpaw import GPAW
from gpaw.utilities.ps2ae import PS2AE
from gpaw.zero_field_splitting import WaveFunctions, zfs1, zfs


def test_zfs_o2(gpw_files):
    calc = GPAW(gpw_files['o2_pw_wfs'])
    D1 = zfs(calc) * 1e6  # ueV
    print(D1)
    assert D1 == pytest.approx(np.diag([117, -59, -59]), abs=1)

    # Poor man's PAW correction:
    converter = PS2AE(calc)
    psit_nR = np.array([converter.get_wave_function(n, ae=True) * bohr**1.5
                        for n in [5, 6]])
    wf2 = WaveFunctions(psit_nR, {}, 0, calc.setups, converter.gd)
    cc = SimpleNamespace(add=lambda a, b: None)
    D2 = zfs1(wf2, wf2, cc) * 1e6  # uev
    print(D2 - D1)
    assert D2 - D1 == pytest.approx(np.diag([14, -7, -7]), abs=1)



"""Calculate the excitation energy of NaCl by an RSF using IVOs."""
from ase.build import molecule
from ase.units import Hartree
from gpaw import GPAW, setup_paths
from gpaw.mpi import world
from gpaw.occupations import FermiDirac
from gpaw.test import equal, gen
from gpaw.eigensolvers import RMMDIIS
from gpaw.cluster import Cluster
from gpaw.lrtddft import LrTDDFT

h = 0.3  # Gridspacing
e_singlet = 4.3
e_singlet_lr = 4.3

if setup_paths[0] != '.':
    setup_paths.insert(0, '.')

gen('Na', xcname='PBE', scalarrel=True, exx=True, yukawa_gamma=0.40)
gen('Cl', xcname='PBE', scalarrel=True, exx=True, yukawa_gamma=0.40)

c = {'energy': 0.005, 'eigenstates': 1e-2, 'density': 1e-2}
mol = Cluster(molecule('NaCl'))
mol.minimal_box(5.0, h=h)
calc = GPAW(txt='NaCl.txt', xc='LCY-PBE:omega=0.40:excitation=singlet',
            eigensolver=RMMDIIS(), h=h, occupations=FermiDirac(width=0.0),
            spinpol=False, convergence=c)
mol.calc = calc
mol.get_potential_energy()
(eps_homo, eps_lumo) = calc.get_homo_lumo()
e_ex = eps_lumo - eps_homo
equal(e_singlet, e_ex, 0.15)
calc.write('NaCl.gpw')

lr = LrTDDFT(calc, txt='LCY_TDDFT_NaCl.log',
             restrict={'istart': 6, 'jend': 7}, nspins=2)
lr.write('LCY_TDDFT_NaCl.ex.gz')
if world.rank == 0:
    lr2 = LrTDDFT.read('LCY_TDDFT_NaCl.ex.gz')
    lr2.diagonalize()
    ex_lr = lr2[1].get_energy() * Hartree
    equal(e_singlet_lr, e_singlet, 0.05)
